from __future__ import annotations

from collections import Counter
from enum import Enum
from typing import TYPE_CHECKING

from stereomolgraph.algorithms.color_refine import color_refine_mg
from stereomolgraph.algorithms.isomorphism import vf2pp_all_isomorphisms
from stereomolgraph.coords import BondsFromDistance
from stereomolgraph.graph2rdmol import mol_graph_to_rdmol, set_crg_bond_orders
from stereomolgraph.graphs.mg import AtomId, Bond, MolGraph

if TYPE_CHECKING:
    
    # Self is included in typing from 3.11

    from collections.abc import Iterator
    from typing import Any, Self

    from rdkit import Chem

    from stereomolgraph.coords import Geometry


class Change(Enum):
    FORMED = "formed"
    FLEETING = "fleeting"
    BROKEN = "broken"

    def __repr__(self) -> str:
        return self.name


class CondensedReactionGraph(MolGraph):
    """
    Graph representing a reaction. Atoms are nodes and (potentially changing)
    bonds are edges. Every node has to have an attribute "atom_type" of type
    Element. Edges can have an attribute "reaction" of type Change.
    This is used to represent the change in connectivity during the reaction.

    Two graphs are equal, iff. they are isomporhic and of the same type.
    """
    __slots__: tuple[str, ...] = tuple()
    _atom_attrs: dict[AtomId, dict[str, Any]]
    _neighbors: dict[AtomId, set[AtomId]]
    _bond_attrs: dict[Bond, dict[str, Any]]

    def __hash__(self) -> int:
        r_color_dict = color_refine_mg(self.reactant())
        p_color_dict = color_refine_mg(self.product())
        ts_colors = color_refine_mg(self)
        color_dict = {a: (r_color_dict[a], ts_colors[a], p_color_dict[a])
                      for a in self.atoms}
        return hash(frozenset(Counter(color_dict.values()).items()))

    def add_bond(self, atom1: int, atom2: int, **attr:Any):
        """
        Adds a bond between atom1 and atom2.

        :param atom1: id of atom1
        :param atom2:   id of atom2
        """
        if "reaction" in attr and not isinstance(
            attr.get("reaction"), Change
        ):
            raise TypeError("reaction bond has to have reaction attribute")
        super().add_bond(atom1, atom2, **attr)

    def set_bond_attribute(
        self, atom1: int, atom2: int, attr: str, value: Any
    ):
        """
        sets the Attribute of the bond between Atom1 and Atom2.

        :param atom1: Atom1
        :param atom2: Atom2
        :param attr: Attribute
        :param value: Value
        """
        if attr == "reaction" and not isinstance(value, Change):
            raise ValueError("reaction bond has to have reaction attribute")
        super().set_bond_attribute(atom1, atom2, attr, value)

    def add_formed_bond(self, atom1: int, atom2: int, **attr: Any):
        """
        Adds a bond between atom1 and atom2 with reaction attribute
        set to FORMED.

        :param atom1: Atom1
        :param atom2: Atom2
        """

        attr["reaction"] = Change.FORMED
        if atom1 in self._atom_attrs and atom2 in self._atom_attrs:
            self.add_bond(atom1, atom2, **attr)
        else:
            raise ValueError("Atoms have to be in the graph")

    def add_broken_bond(self, atom1: int, atom2: int, **attr: Any):
        """
        Adds a bond between atom1 and atom2 with reaction attribute
        set to BROKEN.

        :param atom1: Atom1
        :param atom2: Atom2
        """
        if atom1 in self._atom_attrs and atom2 in self._atom_attrs:
            self.add_bond(atom1, atom2, reaction=Change.BROKEN, **attr)
        else:
            raise ValueError("Atoms have to be in the graph")
            

    def get_formed_bonds(self) -> set[Bond]:
        """
        Returns all bonds that are formed during the reaction

        :return: formed bonds
        """
        f_bonds: set[Bond] = set()
        for bond in self.bonds:
            atom1, atom2 = bond
            if self.get_bond_attribute(atom1, atom2, "reaction") == Change.FORMED:
                f_bonds.add(bond)
        return f_bonds

    def get_broken_bonds(self) -> set[Bond]:
        """
        Returns all bonds that are broken during the reaction

        :return: broken bonds
        """
        b_bonds: set[Bond] = set()
        for bond in self.bonds:
            atom1, atom2 = bond
            if self.get_bond_attribute(atom1, atom2, "reaction") == Change.BROKEN:
                b_bonds.add(bond)
        return b_bonds
        
    def active_atoms(self, additional_layer: int = 0) -> set[int]:
        """
        Atoms involved in the reaction with additional layers of atoms
        in the neighborhood.

        :param additional_layer: Number of additional layers of atoms to
                                 include, defaults to 0
        :return: Atoms involved in the reaction
        """
        active_atoms: set[int] = set()
        for bond in self.get_formed_bonds() | self.get_broken_bonds():
            active_atoms.update(bond)
        for _ in range(additional_layer):
            for atom in active_atoms.copy():
                active_atoms.update(self._neighbors[atom])
        return active_atoms

    def _to_rdmol(
        self,
        generate_bond_orders:bool=False,
        allow_charged_fragments:bool=False,
        charge:int=0
    ) -> tuple[Chem.rdchem.RWMol, dict[int, int]]:
        mol, idx_map_num_dict = mol_graph_to_rdmol(graph=self,
                                                    generate_bond_orders=False,
                                                    allow_charged_fragments=allow_charged_fragments,
                                                    charge=0)
        set_crg_bond_orders(graph=self, mol=mol, idx_map_num_dict=idx_map_num_dict)
        return mol, idx_map_num_dict

    def to_rdmol(
        self,
        generate_bond_orders:bool=False,
        allow_charged_fragments:bool = False,
        charge:int=0
    ) -> Chem.rdchem.Mol:
        raise NotImplementedError(
            "Rdkit is not able to represent "
            "reactions as condensed reaction graphs."
        )

    def reactant(self, keep_attributes:bool=True) -> MolGraph:
        """Reactant of the reaction

        Creates the reactant of the reaction.
        Formed bonds are not present in the reactant.

        :param keep_attributes: attributes on atoms and bonds to be kept,
                                defaults to True
        :return: Reactant of the reaction
        """
        product = MolGraph()
        for atom in self.atoms:
            if keep_attributes is True:
                attrs = self._atom_attrs[atom]
            else:
                attrs = {
                    "atom_type": self._atom_attrs[atom]["atom_type"]
                }
            product.add_atom(atom, **attrs)
        for bond in self.bonds:
            bond_reaction = self._bond_attrs[bond].get("reaction", None)
            if (
                bond_reaction is None or bond_reaction == Change.BROKEN
            ):
                if keep_attributes is True:
                    attrs = self._bond_attrs[bond].copy()
                    attrs.pop("reaction", None)
                else:
                    attrs = {}
                product.add_bond(*bond, **attrs)
        return product

    def product(self, keep_attributes:bool=True) -> MolGraph:
        """Product of the reaction

        Creates the product of the reaction.
        Broken bonds are not present in the product.

        :param keep_attributes: attributes on atoms and bonds to be kept,
                                defaults to True
        :return: Product of the reaction
        """
        product = MolGraph()
        for atom in self.atoms:
            if keep_attributes is True:
                attrs = self._atom_attrs[atom]
            else:
                attrs = {
                    "atom_type": self._atom_attrs[atom]["atom_type"]
                }
            product.add_atom(atom, **attrs)
        for bond in self.bonds:
            bond_reaction = self._bond_attrs[bond].get("reaction", None)
            if (
                bond_reaction is None or bond_reaction == Change.FORMED
            ):
                if keep_attributes is True:
                    attrs = self._bond_attrs[bond].copy()
                    attrs.pop("reaction", None)
                else:
                    attrs = {}
                product.add_bond(*bond, **attrs)
        return product

    def reverse_reaction(self) -> Self:
        """Creates the reaction in the opposite direction.

        Broken bonds are turned into formed bonds and the other way around.

        :return: Reversed reaction
        """
        rev_reac = self.copy()
        for bond in self.bonds:
            bond_reaction = self._bond_attrs[bond].get("reaction", None)
            if bond_reaction == Change.FORMED:
                rev_reac.add_broken_bond(*bond)
            elif bond_reaction == Change.BROKEN:
                rev_reac.add_formed_bond(*bond)

        return rev_reac

    @classmethod
    def from_reactant_and_product_graph(
        cls, reactant_graph: MolGraph, product_graph: MolGraph
    ) -> Self:
        """Creates a CondensedReactionGraph from reactant and product MolGraphs

        CondensedReactionGraph  is constructed from bond changes from reactant
        to the product. The atoms order and atom types of the reactant and
        product have to be the same.

        :param reactant_graph: reactant of the reaction
        :param product_graph: product of the reaction
        :return: CondensedReactionGraph
        """

        if set(reactant_graph.atoms) != set(product_graph.atoms):
            raise ValueError("reactant and product have different atoms1")
        for atom in reactant_graph.atoms:
            if reactant_graph.get_atom_attribute(
                atom, "atom_type"
            ) != product_graph.get_atom_attribute(atom, "atom_type"):
                raise ValueError(
                    "reactant and product have different atom types"
                )

        crg = cls()

        atoms = {*reactant_graph.atoms, *product_graph.atoms}
        for atom in atoms:
            crg.add_atom(
                atom,
                atom_type=reactant_graph.get_atom_type(atom),
            )

        bonds = {
            *[tuple(sorted(bond)) for bond in reactant_graph.bonds],
            *[tuple(sorted(bond)) for bond in product_graph.bonds],
        }

        for bond in bonds:
            if reactant_graph.has_bond(*bond) and product_graph.has_bond(
                *bond
            ):
                crg.add_bond(*bond)
            elif reactant_graph.has_bond(*bond):
                crg.add_bond(*bond, reaction=Change.BROKEN)
            elif product_graph.has_bond(*bond):
                crg.add_bond(*bond, reaction=Change.FORMED)
        return crg

    @classmethod
    def from_reactant_and_product_geometry(
        cls,
        reactant_geo: Geometry,
        product_geo: Geometry,
        switching_function: BondsFromDistance = BondsFromDistance(),
    ) -> Self:
        """Creates a CondensedReactionGraph from reactant
        and product Geometries.


        CondensedReactionGraph  is constructed from bond changes from reactant
        to the product. The atoms order and atom types of the reactant and
        product have to be the same. The switching function is used to
        determine the connectivity of the atoms.

        :param reactant_geo: geometry of the reactant
        :param product_geo: geometry of the product
        :param switching_function: function to define the connectivity
                                   from geometry,
                                   defaults to StepSwitchingFunction()
        :return: CondensedReactionGraph
        """

        reactant = MolGraph.from_geometry(reactant_geo, switching_function)
        product = MolGraph.from_geometry(product_geo, switching_function)
        return cls.from_reactant_and_product_graph(reactant, product)

    def get_isomorphic_mappings(self, other: MolGraph) -> Iterator[dict[int, int]]:
        """Isomorphic mappings between "self" and "other".

        Generates all isomorphic mappings between "other" and "self".
        All atoms and bonds have to be present in both graphs.

        :param other: Other Graph to compare with
        :return: Mappings from the atoms of self onto the atoms of other
        :raises TypeError: Not defined for objects different types
        """
        return vf2pp_all_isomorphisms(
            self,
            other,
            color_refine=True, # TODO: implement color refinement
            stereo=False,
            stereo_change=False,
            subgraph=False,
            # labels=["reaction"],
        )






# ruff: noqa: F401
"""Simple access of key classes MolGraph, StereoMolGraph,
CondensedReactionGraph and StereoCondensedReactionGraph"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from stereomolgraph.graphs.crg import CondensedReactionGraph
    from stereomolgraph.graphs.mg import AtomId, Bond, MolGraph
    from stereomolgraph.graphs.scrg import StereoCondensedReactionGraph
    from stereomolgraph.graphs.smg import StereoMolGraph


def __getattr__(name: str):
    match name:
        case "AtomId":
            from stereomolgraph.graphs.mg import AtomId

            return AtomId
        case "Bond":
            from stereomolgraph.graphs.mg import Bond

            return Bond
        case "MolGraph":
            from stereomolgraph.graphs.mg import MolGraph

            return MolGraph
        case "StereoMolGraph":
            from stereomolgraph.graphs.smg import StereoMolGraph

            return StereoMolGraph
        case "CondensedReactionGraph":
            from stereomolgraph.graphs.crg import CondensedReactionGraph

            return CondensedReactionGraph
        case "StereoCondensedReactionGraph":
            from stereomolgraph.graphs.scrg import StereoCondensedReactionGraph

            return StereoCondensedReactionGraph

        case _:
            raise AttributeError(f"module has no attribute {name}")



from __future__ import annotations

import sys

from collections import defaultdict, Counter
from copy import deepcopy
from types import MappingProxyType
from typing import TYPE_CHECKING, Generic

from stereomolgraph.algorithms.color_refine import color_refine_mg
from stereomolgraph.algorithms.isomorphism import vf2pp_all_isomorphisms
from stereomolgraph.coords import BondsFromDistance
from stereomolgraph.graph2rdmol import set_crg_bond_orders
from stereomolgraph.graphs.crg import CondensedReactionGraph, Change
from stereomolgraph.graphs.mg import AtomId, Bond, MolGraph
from stereomolgraph.graphs.smg import StereoMolGraph
from stereomolgraph.stereodescriptors import (
    AtomStereo,
    BondStereo,
    Stereo,
    Tetrahedral,
)

if TYPE_CHECKING:

    import sys
    from collections.abc import Iterable, Iterator, Mapping
    from typing import Optional
    
    from rdkit import Chem

    from stereomolgraph.coords import Geometry
       
# Self is included in typing from 3.11
if sys.version_info >= (3, 11):
    from typing import Self, TypeVar
else:
    from typing_extensions import Self, TypeVar

S = TypeVar("S", bound="Stereo", contravariant=True)


class ChangeDict(dict[Change, None | S], Generic[S]):
    def __missing__(self, key: Change) -> None:
        if key in Change:
            return None
        else:
            raise KeyError(f"{key} not in {self.__class__.__name__}")
    

class StereoCondensedReactionGraph(StereoMolGraph, CondensedReactionGraph):
    """
    :class:`CondenedReactionGraph` with the ability to store stereochemistry
    information for atoms and (potentially changing) bonds.
    """

    __slots__ = ("_atom_stereo_change", "_bond_stereo_change")
    _atom_stereo_change: defaultdict[AtomId, ChangeDict[AtomStereo]]
    _bond_stereo_change: defaultdict[Bond, ChangeDict[BondStereo]]

    def __init__(self, mol_graph: Optional[MolGraph] = None):
        super().__init__(mol_graph)
        self._atom_stereo_change = defaultdict(ChangeDict[AtomStereo])
        self._bond_stereo_change = defaultdict(ChangeDict[BondStereo])
        
        if mol_graph and isinstance(mol_graph, StereoCondensedReactionGraph):
            self._atom_stereo_change.update(mol_graph._atom_stereo_change)
            self._bond_stereo_change.update(mol_graph._bond_stereo_change)
    
    def __hash__(self) -> int:
        # Get molecular graphs and their colorings
        r = self.reactant()
        p = self.product()
        r_colors = color_refine_mg(r)
        p_colors = color_refine_mg(p)
        ts_colors = color_refine_mg(self)
        colors = {a : hash((r_colors[a], p_colors[a], ts_colors[a]))
                  for a in self.atoms}
        
        # Helper function to compute stereo hashes
        def stereo_hash(stereo):
            return hash(stereo.__class__(
                tuple(colors[a] for a in stereo.atoms
            ), stereo.parity))
        
        # Compute atom and bond stereo hashes
        atom_stereo = {a: stereo_hash(s) for a, s in self.atom_stereo.items()}
        bond_stereo = {b: stereo_hash(s) for b, s in self.bond_stereo.items()}
        
        # Compute stereo change hashes
        def stereo_change_hash(change_dict):
            return tuple(
                (change, stereo_hash(s))
                for change, s in change_dict.items()
                if s is not None
            )
        
        atom_stereo_change = {
            a: hash(stereo_change_hash(change_dict))
            for a, change_dict in self._atom_stereo_change.items()
        }
        
        bond_stereo_change = {
            b: hash(stereo_change_hash(change_dict))
            for b, change_dict in self._bond_stereo_change.items()
        }
        
        # Combine all hashes
        components = (
            colors,
            atom_stereo,
            bond_stereo,
            atom_stereo_change,
            bond_stereo_change
        )
        
        return hash(tuple(
            hash(frozenset(Counter(d.values()).items()))
            for d in components
        ))

    @property
    def atom_stereo_changes(self) -> Mapping[AtomId, ChangeDict[AtomStereo]]:
        return MappingProxyType(self._atom_stereo_change)

    @property
    def bond_stereo_changes(self) -> Mapping[Bond, ChangeDict[BondStereo]]:
        return MappingProxyType(self._bond_stereo_change)

    def get_atom_stereo_change(
        self, atom: int
    ) -> None | Mapping[Change, AtomStereo | None]:
        
        if atom in self._atom_attrs:
            if atom in self._atom_stereo_change:
                return MappingProxyType(self._atom_stereo_change[atom])
            else:
                return None
        else:
            raise ValueError(f"Atom {atom} not in graph")

    def get_bond_stereo_change(
        self, bond: Iterable[int]
        ) -> None | Mapping[Change, BondStereo | None]:
        
        bond = Bond(bond)
        if bond in self._bond_attrs:
            if bond in self._bond_stereo_change:
                return MappingProxyType(self._bond_stereo_change[bond])
            else:
                return None
        else:
            raise ValueError(f"Bond {bond} not in graph")
        
    def set_atom_stereo_change(
        self,
        *,
        broken: Optional[AtomStereo] = None,
        fleeting: Optional[AtomStereo] = None,
        formed: Optional[AtomStereo] = None,
    ):
        atoms: set[int] = set()
        for stereo in (broken, fleeting, formed):
            if stereo is not None:
                atoms.add(stereo.central_atom)
        if len(atoms) != 1:
            raise ValueError("Provide stereo information for one atom only")

        if (atom := atoms.pop()) not in self._atom_attrs:
            raise ValueError(f"Atom {atom} not in graph")
        
        self._atom_stereo_change[atom] = ChangeDict[AtomStereo]()
        for stereo_change, atom_stereo in {
            Change.BROKEN: broken,
            Change.FLEETING: fleeting,
            Change.FORMED: formed,
        }.items():
            if atom_stereo:
                self._atom_stereo_change[atom][stereo_change] = atom_stereo

    def set_bond_stereo_change(
        self,
        *,
        broken: Optional[BondStereo] = None,
        fleeting: Optional[BondStereo] = None,
        formed: Optional[BondStereo] = None,

    ):
        bonds: set[frozenset[int]] = set()
        for stereo in (broken, fleeting, formed):
            if stereo is not None:
                bonds.add(Bond(stereo.bond))
        if len(bonds) != 1:
            raise ValueError("Provide stereo information for one atom only")
        if (bond := bonds.pop()) not in self._bond_attrs:
            raise ValueError(f"Bond {bond} not in graph")
        
        self._bond_stereo_change[bond] = ChangeDict[BondStereo]()
        for stereo_change, bond_stereo in {
            Change.BROKEN: broken,
            Change.FORMED: formed,
            Change.FLEETING: fleeting
        }.items():
            if bond_stereo:
                self._bond_stereo_change[bond][stereo_change] = bond_stereo

    def delete_atom_stereo_change(
        self, atom: AtomId, stereo_change: Optional[Change] = None
    ):
        if stereo_change is None:
            del self._atom_stereo_change[atom]
        else:
            del self._atom_stereo_change[atom][stereo_change]

    def delete_bond_stereo_change(
        self, bond: Iterable[AtomId],
        stereo_change: Optional[Change] = None
    ):
        bond = Bond(bond)
        if stereo_change is None:
            del self._bond_stereo_change[bond]
        else:
            del self._bond_stereo_change[bond][stereo_change]

    def active_atoms(self, additional_layer: int = 0) -> set[AtomId]:
        """
        Atoms involved in the reaction with additional layers of atoms
        in the neighborhood.

        :param additional_layer: Number of additional layers of atoms to
                                 include, defaults to 0
        :return: Atoms involved in the reaction
        """
        active_atoms: set[int] = set()

        for bond in self.get_formed_bonds() | self.get_broken_bonds():
            active_atoms.update(bond)
            
        for _atom, stereo_change in self.atom_stereo_changes.items():
            for _change, stereo in stereo_change.items():
                if stereo is not None:
                    active_atoms.update(stereo.atoms)
        for _bond, stereo_change in self.bond_stereo_changes.items():
            for _change, stereo in stereo_change.items():
                if stereo is not None:
                    active_atoms.update(stereo.atoms)

        for _ in range(additional_layer):
            for atom in active_atoms.copy():
                active_atoms.update(self.bonded_to(atom))
        return active_atoms

    def copy(self) -> Self:
        """
        :return: returns a copy of self
        """
        new_graph = super().copy()
        new_graph._atom_stereo_change = deepcopy(self._atom_stereo_change)
        new_graph._bond_stereo_change = deepcopy(self._bond_stereo_change)
        return new_graph

    def relabel_atoms(
        self, mapping: dict[AtomId, AtomId], copy: bool = True
    ) -> Self:
        """
        Relabels the atoms of the graph and the chiral information accordingly

        :param mapping: Mapping of old atom ids to new atom ids
        :param copy: If the graph should be copied before relabeling,
                     defaults to True
        :return: Returns the relabeled graph or None if copy is False
        """
        relabeled_scrg = self.__class__(
            super().relabel_atoms(mapping, copy=copy)
        )

        atom_stereo_change:defaultdict[AtomId, ChangeDict[AtomStereo]] = defaultdict(ChangeDict[AtomStereo])
        
        for atom, stereo_change_dict in self._atom_stereo_change.items():
            for stereo_change, atom_stereo in stereo_change_dict.items():
                if atom_stereo is None:
                    continue
                new_stereo = atom_stereo.__class__(tuple(mapping.get(atom, atom) for atom in atom_stereo.atoms),
                        atom_stereo.parity)
                atom_stereo_change[mapping[atom]][stereo_change] = new_stereo

        bond_stereo_change: defaultdict[Bond, ChangeDict[BondStereo]] = defaultdict(ChangeDict[BondStereo])
        
        for bond, stereo_change_dict in self._bond_stereo_change.items():
            for stereo_change, bond_stereo in stereo_change_dict.items():
                if bond_stereo is None:
                    continue
                new_bond = Bond(mapping[a] for a in bond)
                new_stereo = bond_stereo.__class__(
                    tuple(mapping.get(atom, atom)
                                  for atom in bond_stereo.atoms),
                    bond_stereo.parity,
                )
                bond_stereo_change[new_bond][stereo_change] = new_stereo
            
        if copy is True:
            relabeled_scrg._atom_stereo_change = atom_stereo_change
            relabeled_scrg._bond_stereo_change = bond_stereo_change
        else:
            self._atom_stereo_change = atom_stereo_change
            self._bond_stereo_change = bond_stereo_change

        return relabeled_scrg

    def reactant(self, keep_attributes:bool=True) -> StereoMolGraph:
        """
        Returns the reactant of the reaction

        :param keep_attributes: If attributes should be kept , defaults to True
        :return: reactant
        """

        reactant = StereoMolGraph(
            super().reactant(keep_attributes=keep_attributes)
        )
        reactant._atom_stereo = deepcopy(self._atom_stereo)
        reactant._bond_stereo = deepcopy(self._bond_stereo)

        for atom, change_dict in self._atom_stereo_change.items():
            if stereo := change_dict[Change.BROKEN]:
                reactant._atom_stereo[atom] = stereo
                

        for _bond, change_dict in self._bond_stereo_change.items():
            if stereo := change_dict[Change.BROKEN]:
                #reactant._bond_stereo[bond] = stereo
                reactant.set_bond_stereo(stereo)

        return reactant

    def product(self, keep_attributes:bool=True) -> StereoMolGraph:
        """
        Returns the product of the reaction

        :param keep_attributes: If attributes should be kept, defaults to True
        :return: product
        """
        product = StereoMolGraph(
            super().product(keep_attributes=keep_attributes)
        )
        product._atom_stereo = deepcopy(self._atom_stereo)
        product._bond_stereo = deepcopy(self._bond_stereo)

        for _atom, change_dict in self._atom_stereo_change.items():
            if stereo := change_dict[Change.FORMED]:
                product.set_atom_stereo(stereo)

        for _bond, change_dict in self._bond_stereo_change.items():
            if stereo := change_dict[Change.FORMED]:
                product.set_bond_stereo(stereo)

        return product

    def reverse_reaction(self) -> Self:
        """Creates the reaction in the opposite direction.

        Broken bonds and stereochemistry changes are turned into formed
        and the other way around.

        :return: Reversed reaction
        """
        rev_reac = super().reverse_reaction()
        
        for _atom, atom_change_dict in rev_reac._atom_stereo_change.items():
            new_atom_change_dict = {"fleeting": atom_change_dict[Change.FLEETING],
                               "broken": atom_change_dict[Change.FORMED],
                               "formed": atom_change_dict[Change.BROKEN]}

            rev_reac.set_atom_stereo_change(**new_atom_change_dict)

        for _bond, bond_change_dict in rev_reac._bond_stereo_change.items():
            new_bond_change_dict = {"fleeting": bond_change_dict[Change.FLEETING],
                               "broken": bond_change_dict[Change.FORMED],
                               "formed": bond_change_dict[Change.BROKEN]}
            rev_reac.set_bond_stereo_change(**new_bond_change_dict)
            
        return rev_reac

    def enantiomer(self) -> Self:
        """
        Creates the enantiomer of the StereoCondensedReactionGraph by inversion
        of all chiral stereochemistries. The result can be identical to the
        molecule itself if the molecule is not chiral.

        :return: Enantiomer
        """
        enantiomer = super().enantiomer()
        for atom in self.atoms:
            stereo_change = self.get_atom_stereo_change(atom=atom)
            if stereo_change is not None:
                stereo_change_inverted = {
                    change.value: stereo.invert() if stereo else None
                    for change, stereo in stereo_change.items()
                }
                enantiomer.set_atom_stereo_change(
                    **stereo_change_inverted
                )
        return enantiomer

    def _to_rdmol(
        self,
        generate_bond_orders:bool=False,
        allow_charged_fragments:bool=False,
        charge:int=0
    ) -> tuple[Chem.rdchem.RWMol, dict[int, int]]:
        
        ts_smg = StereoMolGraph(self) # bond change is now just a bond
        
        for _atom, stereo_change_dict in self.atom_stereo_changes.items():
            atom_stereo = next((stereo for stereo_change in (Change.FLEETING,
                                                             Change.BROKEN,
                                                             Change.FORMED)
                                if (stereo := stereo_change_dict[stereo_change]) is not None), None)
            if atom_stereo:
                ts_smg.set_atom_stereo(atom_stereo)

        for _bond, stereo_change_dict in self.bond_stereo_changes.items():
            bond_stereo = next((stereo for stereo_change in (Change.FLEETING,
                                                             Change.BROKEN,
                                                             Change.FORMED)
                                if (stereo := stereo_change_dict[stereo_change]) is not None), None)
            if bond_stereo:
                ts_smg.set_bond_stereo(bond_stereo)

        mol, idx_map_num_dict = ts_smg._to_rdmol(
            generate_bond_orders=False,
            allow_charged_fragments=allow_charged_fragments,
            charge=charge)
        if generate_bond_orders:
            mol = set_crg_bond_orders(graph=self,
                                 mol=mol,
                                 charge=charge,
                                 idx_map_num_dict=idx_map_num_dict)
        return mol, idx_map_num_dict

    @classmethod
    def compose(cls, mol_graphs: Iterable[MolGraph]) -> Self:
        """Creates a MolGraph object from a list of MolGraph objects

        :param mol_graphs: list of MolGraph objects
        :return: Returns Combined MolGraph
        """
        graph = cls(super().compose(mol_graphs))
        for mol_graph in mol_graphs:
            graph._atom_stereo_change.update(cls(mol_graph)._atom_stereo_change)
            graph._bond_stereo_change.update(cls(mol_graph)._bond_stereo_change)
        return graph

    @classmethod
    def from_reactant_and_product_graph(
        cls,
        reactant_graph: MolGraph,
        product_graph: MolGraph,
    ) -> Self:
        """Creates a StereoCondensedReactionGraph from reactant and product
        StereoMolGraphs.

        StereoCondensedReactionGraph  is constructed from bond changes from
        reactant to the product. The atoms order and atom types of the reactant
        and product have to be the same.

        :param reactant_graph: reactant of the reaction
        :param product_graph: product of the reaction
        :return: StereoCondensedReactionGraph
        """
        crg = super().from_reactant_and_product_graph(
            reactant_graph, product_graph
        )
        scrg = cls(crg)

        reactant_graph = StereoMolGraph(reactant_graph)
        product_graph = StereoMolGraph(product_graph)

        all_stereo_atoms = set(reactant_graph._atom_stereo) | set(
            product_graph._atom_stereo
        )

        for atom in all_stereo_atoms:
            r_stereo = reactant_graph.get_atom_stereo(atom)
            p_stereo = product_graph.get_atom_stereo(atom)

            if (
                r_stereo is not None
                and p_stereo is not None
                and r_stereo == p_stereo
            ):
                scrg.set_atom_stereo(r_stereo)

            elif r_stereo is None and p_stereo is not None:
                scrg.set_atom_stereo_change(formed=p_stereo)

            elif p_stereo is None and r_stereo is not None:
                scrg.set_atom_stereo_change(broken=r_stereo)

            elif (
                r_stereo is not None
                and p_stereo is not None
                and r_stereo != p_stereo
            ):
                scrg.set_atom_stereo_change(
                    formed=p_stereo, broken=r_stereo
                )

        all_stereo_bonds = set(reactant_graph._bond_stereo) | set(
            product_graph._bond_stereo
        )

        for bond in all_stereo_bonds:
            r_stereo = reactant_graph.get_bond_stereo(bond)
            p_stereo = product_graph.get_bond_stereo(bond)

            if (
                r_stereo is not None
                and p_stereo is not None
                and r_stereo == p_stereo
            ):
                scrg.set_bond_stereo(r_stereo)

            elif r_stereo is None and p_stereo is not None:
                scrg.set_bond_stereo_change(formed=p_stereo)

            elif p_stereo is None and r_stereo is not None:
                scrg.set_bond_stereo_change(broken=r_stereo)

            elif (
                r_stereo is not None
                and p_stereo is not None
                and r_stereo != p_stereo
            ):
                scrg.set_bond_stereo_change(
                    formed=p_stereo, broken=r_stereo
                )

        for atom in scrg.atoms:
            isnon = (
                    scrg.get_atom_stereo_change(atom) is None
                    or (change_d := scrg.get_atom_stereo_change(atom)) is None
                    or all (stereo is None
                        for stereo in change_d.values()))
            if (isnon
                and reactant_graph.get_atom_stereo(atom) is None
                and product_graph.get_atom_stereo(atom) is None
                and len(scrg.bonded_to(atom)) == 4
            ):
                t_atoms = (atom, *scrg.bonded_to(atom))
                assert len(t_atoms) == 5
                scrg.set_atom_stereo_change(
                    broken=Tetrahedral(t_atoms, None),
                    formed=Tetrahedral(t_atoms, None),
                )
                # TODO: add someting for 5 or more substituents

        return scrg

    @classmethod
    def from_reactant_and_product_geometry(
        cls,
        reactant_geo: Geometry,
        product_geo: Geometry,
        switching_function: BondsFromDistance = BondsFromDistance(),
    ) -> Self:
        """Creates a StereoCondensedReactionGraph from reactant and product
        Geometries.

        StereoCondensedReactionGraph  is constructed from bond changes from
        reactant to the product. The atoms order and atom types of the reactant
        and product have to be the same. The switching function is used to
        determine the connectivity of the atoms.

        :param reactant_geo: geometry of the reactant
        :param product_geo: geometry of the product
        :param switching_function: function to define the connectivity from
                                   geometry,
                                   defaults to StepSwitchingFunction()
        :return: StereoCondensedReactionGraph
        """
        reactant_graph = StereoMolGraph.from_geometry(
            reactant_geo, switching_function
        )
        product_graph = StereoMolGraph.from_geometry(
            product_geo, switching_function
        )

        return cls.from_reactant_and_product_graph(
            reactant_graph=reactant_graph, product_graph=product_graph
        )

    @classmethod
    def from_reactant_product_and_ts_geometry(
        cls,
        reactant_geo: Geometry,
        product_geo: Geometry,
        ts_geo: Geometry,
        switching_function: BondsFromDistance = BondsFromDistance(),
    ) -> Self:
        """Creates a StereoCondensedReactionGraph from reactant, product and
        transition state Geometries.

        StereoCondensedReactionGraph  is constructed from bond changes from
        reactant to the product. The atoms order and atom types of the reactant
        and product have to be the same. The switching function is used to
        determine the connectivity of the atoms. Only the stereo information
        is taken from the transition state geometry.

        :param reactant_geo: geometry of the reactant
        :param product_geo: geometry of the product
        :param ts_geo: geometry of the transition state
        :param switching_function: function to define the connectivity from
                                   geometry,
                                   defaults to StepSwitchingFunction()
        :return: CondensedReactionGraph
        """

        crg = CondensedReactionGraph.from_reactant_and_product_geometry(
            reactant_geo=reactant_geo,
            product_geo=product_geo,
            switching_function=switching_function,
        )

        ts_atom_stereo_graph = StereoMolGraph(crg)
        ts_atom_stereo_graph._set_atom_stereo_from_geometry(ts_geo)
        reactant_atom_stereo_graph = StereoMolGraph.from_geometry(
            geo=reactant_geo, switching_function=switching_function
        )
        product_atom_stereo_graph = StereoMolGraph.from_geometry(
            geo=product_geo, switching_function=switching_function
        )

        scrg = cls.from_reactant_and_product_graph(
            reactant_graph=reactant_atom_stereo_graph,
            product_graph=product_atom_stereo_graph,
        )

        for atom in ts_atom_stereo_graph._atom_stereo:
            r_atom_stereo = reactant_atom_stereo_graph.get_atom_stereo(atom)
            ts_atom_stereo = ts_atom_stereo_graph.get_atom_stereo(atom)
            p_atom_stereo = product_atom_stereo_graph.get_atom_stereo(atom)

    
            if (r_atom_stereo is not None
                and p_atom_stereo is not None
                and ts_atom_stereo is not None
                and r_atom_stereo != p_atom_stereo
                and r_atom_stereo != ts_atom_stereo
                and p_atom_stereo != ts_atom_stereo

            ):
                scrg.set_atom_stereo_change(
                    broken=r_atom_stereo,
                    formed=p_atom_stereo,
                    fleeting=ts_atom_stereo,
                )

        return scrg

    def get_isomorphic_mappings(
        self, other: MolGraph, stereo:bool=True, stereo_change:bool=True
    ) -> Iterator[dict[int, int]]:
        """Isomorphic mappings between "self" and "other".

        Generates all isomorphic mappings between "other" and "self".
        All atoms and bonds have to be present in both graphs.
        The Stereochemistry is preserved in the mappings.

        :param other: Other Graph to compare with
        :return: Mappings from the atoms of self onto the atoms of other
        :raises TypeError: Not defined for objects different types
        """
        return vf2pp_all_isomorphisms(
            self,
            other,
            labels=None,
            color_refine=True, # TODO: implement color refinement
            stereo=stereo,
            stereo_change=stereo_change,
            subgraph=False,
        )

    def get_subgraph_isomorphic_mappings(
        self, other: MolGraph, stereo:bool=True, stereo_change:bool=True
    ) -> Iterator[dict[int, int]]:
        """Subgraph isomorphic mappings from "other" onto "self".

        Generates all node-iduced subgraph isomorphic mappings.
        All atoms of "other" have to be present in "self".
        The bonds of "other" have to be the subset of the bonds of "self"
        relating to the nodes of "other".
        The Stereochemistry is preserved in the mappings.

        :param other: Other Graph to compare with
        :return: Mappings from the atoms of self onto the atoms of other
        :raises TypeError: Not defined for objects different types
        """
        return vf2pp_all_isomorphisms(
            self,
            other,
            labels=None,
            color_refine=False,
            stereo=stereo,
            stereo_change=stereo_change,
            subgraph=True,
        )

    




import tempfile
import textwrap
from pathlib import Path

import numpy as np
import pytest

from stereomolgraph.coords import Geometry, are_planar, angle_from_coords, handedness, pairwise_distances


class TestGeometry:
    @pytest.mark.parametrize(
        "xyz_content, comment, coords",
        [
            (  # with comment
                textwrap.dedent("""\
                    2
                        this is a comment comment with whitespace
                    C                 -3.7    0.02    0.2
                    C                 -3.1   -1.1   -0.2
                """),
                "this is a comment comment with whitespace",
                [[-3.7, 0.02, 0.2], [-3.1, -1.1, -0.2]],
            ),
            (  # with empty lines at end
                textwrap.dedent("""\
                    2

                    C                 -3.7    0.02    0.2
                    C                  0.0    0.0     0.0

                """),
                "",
                [[-3.7, 0.02, 0.2], [0.0, 0.0, 0.0]],
            ),
        ],
        ids=["with comment", "empty lines at end"],
    )
    def test_from_xyz_file(self, xyz_content, comment, coords):
        with tempfile.NamedTemporaryFile("w+", delete=False) as tmp:
            tmp.write(xyz_content)
            tmp.flush()
            fake_path = Path(tmp.name)

            geo = Geometry.from_xyz_file(fake_path)

        np.testing.assert_equal(geo.coords, coords)




class TestGeometryFunctions:
    """Test suite for geometry functions"""
    
    # Test data
    PLANAR_POINTS = np.array([
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
        [0.5, 0.5, 0]  # Perfectly planar
    ], dtype=np.float64) * 100
    
    NON_PLANAR_POINTS = np.array([
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
        [0.5, 0.5, 0.1]  # Slightly out of plane
    ], dtype=np.float64) * 100
    
    RIGHT_ANGLE_POINTS = np.array([
        [1, 0, 0],  # A
        [0, 0, 0],  # B (vertex)
        [0, 1, 0]   # C
    ], dtype=np.float64) * 100
    
    STRAIGHT_ANGLE_POINTS = np.array([
        [1, 0, 0],
        [0, 0, 0],
        [-1, 0, 0]
    ], dtype=np.float64) * 100
    
    CHIRAL_TETRAHEDRON = np.array([
        [0, 0, 0],  # Central atom
        [1, 0, 0],  # Right
        [0, 1, 0],  # Front
        [0, 0, 1]   # Up (R configuration)
    ], dtype=np.float64) * 100
    
    CHIRAL_TETRAHEDRON_S = np.array([
        [0, 0, 0],  # Central atom
        [1, 0, 0],  # Right
        [0, 1, 0],  # Front
        [0, 0, -1]  # Down (S configuration)
    ], dtype=np.float64) * 100
    
    def test_are_planar_non_vectorized(self):
        """Test planar check with single point set"""
        assert are_planar(self.PLANAR_POINTS)
        assert not are_planar(self.NON_PLANAR_POINTS)
        assert are_planar(self.NON_PLANAR_POINTS/100, threshold=1)  # Should pass with larger threshold
        
    @pytest.mark.skip("Not implemented yet")
    def test_are_planar_vectorized(self):
        """Test planar check with multiple point sets"""
        multiple_sets = np.stack([self.PLANAR_POINTS, self.NON_PLANAR_POINTS])
        result = are_planar(multiple_sets)
        assert result.shape == (2,)
        assert result[0] is np.bool_(True)
        assert result[1] is np.bool_(False)
        
    @pytest.mark.skip("Not implemented yet")
    def test_are_planar_edge_cases(self):
        with pytest.raises(AssertionError):
            assert are_planar(self.PLANAR_POINTS[:3])
        
        # Test with colinear points (degenerate case)
        colinear = np.array([
            [0, 0, 0],
            [1, 0, 0],
            [2, 0, 0],
            [3, 0, 0]
        ])
        assert are_planar(colinear)  # Technically planar
        
    def test_angle_from_coords_non_vectorized(self):
        """Test angle calculation with single point set"""
        angle = angle_from_coords(self.RIGHT_ANGLE_POINTS)
        assert np.isclose(angle, 90.0)
        
        angle = angle_from_coords(self.STRAIGHT_ANGLE_POINTS)
        assert np.isclose(angle, 180.0)
        
        # Test degenerate case (zero-length vectors) not included
        degenerate = np.array([
            [0, 0, 0],
            [0, 0, 0],  # All points same
            [0, 0, 0]
        ], dtype=np.float64)
        angle = angle_from_coords(degenerate)
        #assert np.isnan(angle)
        
    def test_angle_from_coords_vectorized(self):
        """Test angle calculation with multiple point sets"""
        multiple_angles = np.stack([
            self.RIGHT_ANGLE_POINTS,
            self.STRAIGHT_ANGLE_POINTS,
            np.array([[1,1,0], [0,0,0], [-1,1,0]])  # 90° angle
        ])
        
        angles = angle_from_coords(multiple_angles)
        assert angles.shape == (3,)
        assert np.allclose(angles, [90.0, 180.0, 90.0])
        
    def test_handedness_non_vectorized(self):
        """Test chirality calculation with single point set"""
        # Test R configuration
        result = handedness(self.CHIRAL_TETRAHEDRON)
        assert result == -1
        
        # Test S configuration
        result = handedness(self.CHIRAL_TETRAHEDRON_S)
        assert result == 1
        
        # Test planar case (should be 0)
        planar = np.vstack([self.PLANAR_POINTS[:3], [0.5, 0.5, 0]])
        result = handedness(planar)
        assert result == 0
        
    def test_handedness_vectorized(self):
        """Test chirality calculation with multiple point sets"""
        multiple_configs = np.stack([
            self.CHIRAL_TETRAHEDRON,
            self.CHIRAL_TETRAHEDRON_S,
            np.vstack([self.PLANAR_POINTS[:3], [0.5, 0.5, 0]])  # Planar
        ])
        
        results = handedness(multiple_configs)
        assert results.shape == (3,)
        assert np.array_equal(results, [-1, 1, 0])
        
    def test_pairwise_distances_non_vectorized(self):
        """Test distance matrix calculation"""
        points = np.array([
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0]
        ], dtype=np.float64)
        
        dist_matrix = pairwise_distances(points)
        expected = np.array([
            [0, 1, 1],
            [1, 0, np.sqrt(2)],
            [1, np.sqrt(2), 0]
        ])
        
        assert dist_matrix.shape == (3, 3)
        assert np.allclose(dist_matrix, expected)
        
    def test_pairwise_distances_vectorized(self):
        """Test batched distance matrix calculation"""
        batch_points = np.stack([
            self.PLANAR_POINTS,
            self.PLANAR_POINTS,
            self.PLANAR_POINTS,
            self.PLANAR_POINTS,
            self.PLANAR_POINTS
        ])
        result = pairwise_distances(self.PLANAR_POINTS)
        vec_results = pairwise_distances(batch_points)

        assert vec_results.shape == (5, 4, 4)
        for r in vec_results:
            assert r.shape == (4, 4)
            assert np.allclose(r, result)

    @pytest.mark.skip(reason="NaN handling not implemented yet")
    def test_all_functions_nan_handling(self):
        """Test how functions handle NaN inputs"""
        nan_points = np.full((4, 3), np.nan)
        
        with pytest.raises(ValueError):
            are_planar(nan_points)
            
        with pytest.raises(ValueError):
            handedness(nan_points[:4])
            
        with pytest.raises(ValueError):
            angle_from_coords(nan_points[:3])
            
        with pytest.raises(ValueError):
            pairwise_distances(nan_points)


from collections import defaultdict
from copy import deepcopy
from itertools import permutations

import numpy as np
import pytest
import rdkit.Chem  # type: ignore

from stereomolgraph import PERIODIC_TABLE as PTOE
from stereomolgraph import (Bond,
                            MolGraph,
                            StereoMolGraph,
                            CondensedReactionGraph,
                            StereoCondensedReactionGraph)
from stereomolgraph.coords import Geometry, are_planar
from stereomolgraph.graphs.crg import Change
from stereomolgraph.graphs.scrg import ChangeDict

from stereomolgraph.stereodescriptors import (
    AtropBond,
    Octahedral,
    PlanarBond,
    Tetrahedral,
    TrigonalBipyramidal,
)

class TestMolGraph:
    _TestClass: type[MolGraph] = MolGraph

    @pytest.fixture
    def enantiomer_graph1(self, enantiomer_geos):
        return self._TestClass.from_geometry(enantiomer_geos[0])

    @pytest.fixture
    def enantiomer_graph2(self, enantiomer_geos):
        return self._TestClass.from_geometry(enantiomer_geos[1])

    @pytest.fixture
    def water_graph(self, water_geo):
        return self._TestClass.from_geometry(water_geo)

    @pytest.fixture
    def empty_mol_graph(self):
        return self._TestClass()

    @pytest.fixture
    def mol_graph(self):
        mol_graph = self._TestClass()
        mol_graph.add_atom(0, atom_type="C")
        mol_graph.add_atom(1, atom_type="H")
        mol_graph.add_atom(2, atom_type="O")
        mol_graph.add_bond(0, 1, bond_order=1)
        return mol_graph

    @pytest.fixture
    def chiral_product_geo1(self, data_path):
        filepath = data_path / "disrot_reaction" / "(Z)-(4S)-3,4-Dichlor-2-pentene.xyz"
        
        return Geometry.from_xyz_file(filepath)

    @pytest.fixture
    def chiral_product_graph1(self, chiral_product_geo1):
        return self._TestClass.from_geometry(chiral_product_geo1)

    @pytest.fixture
    def chiral_product_geo2(self, data_path):
        filepath = data_path / "conrot_reaction/(Z)-(4S)-3,4-Dichlor-2-pentene.xyz"
        return Geometry.from_xyz_file(filepath)

    @pytest.fixture
    def chiral_product_graph2(self, chiral_product_geo2):
        return self._TestClass.from_geometry(chiral_product_geo2)

    @pytest.fixture
    def chiral_reactant_geo(self, data_path):
        filepath = data_path /"conrot_reaction/(2S,3S)-1,1-Dichlor-2,3-dimethylcyclopropane.xyz"
        
        return Geometry.from_xyz_file(filepath)

    @pytest.fixture
    def chiral_reactant_graph(self, chiral_reactant_geo):
        return self._TestClass.from_geometry(chiral_reactant_geo)

    def test_len(self, enantiomer_graph1):
        assert len(enantiomer_graph1) == 8

    def test_add_atom(self, empty_mol_graph, *args, **kwargs):
        empty_mol_graph.add_atom(0, atom_type="H", *args, **kwargs)
        empty_mol_graph.add_atom(1, atom_type=PTOE["C"], *args, **kwargs)
        assert empty_mol_graph

    def test_remove_atom(self, enantiomer_graph1):
        enantiomer_graph1.remove_atom(0)
        assert 0 not in enantiomer_graph1.atoms

    def test_add_bond(self, enantiomer_graph1):
        enantiomer_graph1.add_bond(0, 6)
        assert Bond({0, 6}) in enantiomer_graph1.bonds

    def test_remove_bond(self, water_graph):
        water_graph.remove_bond(0, 1)
        assert Bond({0, 1}) not in water_graph.bonds

    def test_atoms(self, enantiomer_graph1):
        assert set(enantiomer_graph1.atoms) == {0, 1, 2, 3, 4, 5, 6, 7}

    def test_bonds(self, mol_graph):
        assert Bond((0, 1)) in mol_graph.bonds

    def test_get_atom_attribute(self, mol_graph):
        assert mol_graph.get_atom_attribute(1, attr="atom_type") == PTOE["H"]

    def test_get_bond_attribute(self, mol_graph):
        assert mol_graph.get_bond_attribute(0, 1, attr="bond_order") == 1

    def test_get_atom_attributes(self, mol_graph):
        assert mol_graph.get_atom_attributes(1) == {"atom_type": PTOE["H"]}

    def test_get_atoms_with_attributes(self, mol_graph):
        assert mol_graph.atoms_with_attributes == {
            0: {"atom_type": PTOE["C"]},
            1: {"atom_type": PTOE["H"]},
            2: {"atom_type": PTOE["O"]},
        }

    def test_set_atom_attribute(self, mol_graph):
        mol_graph.set_atom_attribute(1, attr="test_attr", value="test")
        assert mol_graph.get_atom_attribute(1, attr="test_attr") == "test"
        mol_graph.set_atom_attribute(1, attr="atom_type", value="He")
        assert mol_graph.get_atom_attribute(1, attr="atom_type") == PTOE["He"]
        with pytest.raises(ValueError):
            mol_graph.set_atom_attribute(1, attr="atom_type", value="test")

    def test_set_bond_attribute(self, mol_graph):
        mol_graph.set_bond_attribute(0, 1, attr="lengh", value="very_long")
        assert mol_graph.get_bond_attribute(0, 1, attr="lengh") == "very_long"
        mol_graph.set_bond_attribute(0, 1, attr="bond_order", value=13)
        assert mol_graph.get_bond_attribute(0, 1, attr="bond_order") == 13

    def test_delete_atom_attribute(self, mol_graph):
        mol_graph.set_atom_attribute(1, attr="test_attr", value="test")
        assert (
            mol_graph.get_atom_attribute(1, attr="test_attr")
            is not None
        )
        mol_graph.delete_atom_attribute(1, attr="test_attr")
        assert (
            mol_graph.get_atom_attribute(1, attr="test_attr")
            is None
        )

    def test_delete_bond_attribute(self, mol_graph):
        mol_graph.delete_bond_attribute(0, 1, attr="bond_order")
        assert (
            mol_graph.get_bond_attribute(0, 1, attr="bond_oder")
            is None
        )

    def test_get_bond_attributes(self, mol_graph):
        assert mol_graph.get_bond_attributes(0, 1) == {"bond_order": 1}

    def test_get_bonds_with_attributes(self, mol_graph):
        assert mol_graph.bonds_with_attributes == {
            Bond({0, 1}): {"bond_order": 1}
        }

    def test_connectivity_matrix(self, mol_graph):
        assert np.array_equal(
            mol_graph.connectivity_matrix(),
            np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]], dtype=int),
        )

    def test_relabel_atoms(self, enantiomer_graph1):
        mapping = {0: 15, 5: 10, 3: 99}
        enantiomer_graph1.relabel_atoms(mapping, copy=False)
        assert all(
            atom in enantiomer_graph1.atoms for atom in mapping.values()
        )
        assert all(
            atom not in enantiomer_graph1.atoms for atom in mapping.keys()
        )

    def test_relabel_atoms_copy(self, enantiomer_graph1):
        mapping = {0: 15, 5: 10, 3: 99}
        new_graph = enantiomer_graph1.relabel_atoms(mapping, copy=True)
        assert all(atom in new_graph.atoms for atom in mapping.values())
        assert all(atom not in new_graph.atoms for atom in mapping.keys())

    #@pytest.mark.skip("Not Implemented")
    def test_connected_components(self, mol_graph):
        assert [i for i in mol_graph.connected_components()] == [{0, 1}, {2}]

    def test_subgraph(self, mol_graph):
        subgraph1 = mol_graph.subgraph([0, 1])
        assert (
            1 in subgraph1.atoms
            and 0 in subgraph1.atoms
            and Bond({0, 1}) in subgraph1.bonds
            and 2 not in subgraph1.atoms
        )
        subgraph2 = mol_graph.subgraph([1, 2])
        assert (
            1 in subgraph2.atoms
            and 2 in subgraph2.atoms
            and Bond({1, 2}) not in subgraph2.bonds
            and 0 not in subgraph2.atoms
        )

    def test_copy(self, enantiomer_graph1):
        copied_mol_graph = enantiomer_graph1.copy()
        assert id(copied_mol_graph) != id(enantiomer_graph1)

    def test_get_isomorphic_mappings(self, water_graph, mol_graph):
        assert [] == [
            i for i in mol_graph.get_isomorphic_mappings(water_graph)
        ]
        
        assert all(
            mapping in ({0: 0, 1: 1, 2: 2}, {0: 0, 2: 1, 1: 2})
            for mapping in (
                i for i in water_graph.get_isomorphic_mappings(water_graph)
            )
        )

    def test_get_isomorphic_mappings_of_enantiomers(
        self, enantiomer_graph1, enantiomer_graph2
    ):
        assert 2 == len(
            list(enantiomer_graph1.get_isomorphic_mappings(enantiomer_graph2))
        )


    def test_get_automorphic_mappings(self, water_graph):
        assert all(
            mapping in ({0: 0, 1: 1, 2: 2}, {0: 0, 2: 1, 1: 2})
            for mapping in (i for i in water_graph.get_isomorphic_mappings(water_graph))
        )

    def test_compose(
        self, water_graph, mol_graph, empty_mol_graph
    ):

        comp_graph = self._TestClass.compose(
            [water_graph, empty_mol_graph]
        )
        assert comp_graph.atom_types == water_graph.atom_types
        assert comp_graph.atoms == water_graph.atoms
        assert comp_graph.bonds == water_graph.bonds

        comp_graph = self._TestClass.compose(
            [water_graph, mol_graph]
        )
        assert comp_graph.atom_types == mol_graph.atom_types
        assert Bond((0, 2)) in comp_graph.bonds
        assert Bond((0, 1)) in comp_graph.bonds
        assert (
            comp_graph.get_bond_attribute(0, 1, "bond_order") == 1
        )

    def test_from_composed_chiral_molgraphs(
        self, chiral_product_graph1, chiral_product_graph2
    ):
        relabel_mapping = {
            atom: atom + chiral_product_graph1.n_atoms
            for atom in chiral_product_graph2.atoms
        }
        chiral_product_graph2.relabel_atoms(relabel_mapping, copy=False)

        combined = self._TestClass.compose(
            [chiral_product_graph1, chiral_product_graph2]
        )

        assert (
            combined.atoms_with_attributes
            == chiral_product_graph1.atoms_with_attributes
            | chiral_product_graph2.atoms_with_attributes
        )
        assert (
            combined.bonds_with_attributes
            == chiral_product_graph1.bonds_with_attributes
            | chiral_product_graph2.bonds_with_attributes
        )

    def test_to_rdmol(self, water_graph):
        rdmol, _ = water_graph._to_rdmol()
        assert [Atom.GetAtomicNum() for Atom in rdmol.GetAtoms()] == [
            atom_type.atomic_nr for atom_type in water_graph.atom_types
        ]
        assert ({Bond(
                (rd_b.GetBeginAtomIdx(), rd_b.GetEndAtomIdx()))
                for rd_b in rdmol.GetBonds()}
                == {Bond(b) for b in water_graph.bonds})
        
    def test_from_rdmol(self):

        rdmol = rdkit.Chem.MolFromSmiles("[C:1]([O:2][C:33]([C:4]([O:5][H:13])"
                                         "([H:111])[H:12])([H:9])[H:10])([H:6])"
                                         "([H:77])[H:8]", sanitize=False)
        mol_graph = self._TestClass.from_rdmol(rdmol, use_atom_map_number=True)

        assert set(mol_graph.atoms) == set((1,2,33,4,5,6,77,8,9,10,111,12,13))
        assert mol_graph.get_atom_attribute(1, "atom_type") == PTOE["C"]
        assert mol_graph.get_atom_attribute(2, "atom_type") == PTOE["O"]
        assert mol_graph.get_atom_attribute(33, "atom_type") == PTOE["C"]
        assert mol_graph.get_atom_attribute(4, "atom_type") == PTOE["C"]
        assert mol_graph.get_atom_attribute(5, "atom_type") == PTOE["O"]
        assert mol_graph.get_atom_attribute(6, "atom_type") == PTOE["H"]
        assert mol_graph.get_atom_attribute(77, "atom_type") == PTOE["H"]
        assert mol_graph.get_atom_attribute(8, "atom_type") == PTOE["H"]
        assert mol_graph.get_atom_attribute(9, "atom_type") == PTOE["H"]
        assert mol_graph.get_atom_attribute(10, "atom_type") == PTOE["H"]
        assert mol_graph.get_atom_attribute(111, "atom_type") == PTOE["H"]
        assert mol_graph.get_atom_attribute(12, "atom_type") == PTOE["H"]
        assert mol_graph.get_atom_attribute(13, "atom_type") == PTOE["H"]

        assert mol_graph.has_bond(1, 2)
        assert mol_graph.has_bond(2, 33)
        assert mol_graph.has_bond(33, 4)
        assert mol_graph.has_bond(4, 5)
        assert mol_graph.has_bond(1, 6)
        assert mol_graph.has_bond(1, 77)
        assert mol_graph.has_bond(1, 8)
        assert mol_graph.has_bond(33, 9)
        assert mol_graph.has_bond(33, 10)
        assert mol_graph.has_bond(4, 111)
        assert mol_graph.has_bond(4, 12)
        assert mol_graph.has_bond(5, 13)

    @pytest.mark.parametrize("inchi", [
    (r"InChI=1S/C3H8O/c1-3(2)4/h3-4H,1-2H3"),
    (r"InChI=1S/C8H10N4O2/c1-10-4-9-6-5(10)7(13)12(3)8(14)11(6)2/h4H,1-3H3")
    ], ids = ["isopropanol",
              "caffeine"],)
    def test_from_rdmol_to_rdmol_not_chiral(self, inchi):
        rdmol = rdkit.Chem.MolFromInchi(inchi, sanitize=False, removeHs=False)
        assert inchi == rdkit.Chem.MolToInchi(rdmol, treatWarningAsError=True) # type: ignore
        
        molgraph = self._TestClass.from_rdmol(rdmol)
        rdmol2, _ = molgraph._to_rdmol(generate_bond_orders=True) 
        # TODO: check why bond orders are needed

        assert inchi == rdkit.Chem.MolToInchi(rdmol2, treatWarningAsError=True) # type: ignore

    def test_equality_relabeled_water(self, water_graph):
        assert water_graph == water_graph.copy()
        assert water_graph == water_graph.relabel_atoms({0: 1, 1: 0, 2: 13})

    def test_equality(
        self,
        chiral_product_graph1,
        chiral_product_graph2,
        chiral_reactant_graph,
    ):
        assert chiral_product_graph1 == chiral_product_graph2
        assert (
            chiral_product_graph1
            != chiral_reactant_graph
            != chiral_product_graph2
        )

    def test_hash_enantiomers(self, enantiomer_graph1, enantiomer_graph2):
        assert (hash(enantiomer_graph1)
                == hash(enantiomer_graph2))

    def test_hash_relabel(self, water_graph):
        relabel_water = water_graph.relabel_atoms(
            {0: 1, 1: 0, 2: 13}, copy=True
        )
        assert (hash(water_graph)
                == hash(relabel_water))


class TestCondensedReactionGraph(TestMolGraph):
    _TestClass: type[CondensedReactionGraph] = CondensedReactionGraph

    @pytest.fixture
    def crg(self):
        crg = self._TestClass()
        crg.add_atom(0, atom_type="C", atom_label="atom_label_C1")
        crg.add_atom(1, atom_type="H")
        crg.add_atom(2, atom_type="O")
        crg.add_atom(3, atom_type="H")
        crg.add_bond(0, 1, reaction=Change.FORMED)
        crg.add_bond(0, 2, bond_label="bond_label_C1O3")
        crg.add_bond(2, 3, reaction=Change.BROKEN)
        return crg

    def test_bonds(self, chiral_reactant_graph, chiral_product_graph1,
                            chiral_reaction_scrg1):
        unchanged_b = {bond for bond
                           in chiral_reaction_scrg1.bonds
                           if bond
                           not in chiral_reaction_scrg1.get_broken_bonds()
                           and bond
                           not in chiral_reaction_scrg1.get_formed_bonds()}

        formed_b = chiral_reaction_scrg1.get_formed_bonds()
        broken_b = chiral_reaction_scrg1.get_broken_bonds()

        assert set(chiral_reactant_graph.bonds) == unchanged_b | broken_b
        assert set(chiral_product_graph1.bonds) == unchanged_b | formed_b

    def test_add_bond_error(self, crg):
        with pytest.raises(TypeError):
            crg.add_bond(0, 1, reaction="test")

    def test_add_bond_with_reaction_attr(self, crg):
        crg.add_bond(0, 1, reaction=Change.FORMED, bond_order=1)
        assert (
            crg.get_bond_attribute(0, 1, attr="reaction") == Change.FORMED
        )
        assert crg.get_bond_attribute(0, 1, attr="bond_order") == 1

    def test_set_bond_attribute_reaction_exception(self, crg):
        with pytest.raises(ValueError):
            crg.set_bond_attribute(0, 1, attr="reaction", value="test")

    def test_set_bond_attribute_reaction(self, crg):
        crg.set_bond_attribute(0, 1, attr="reaction", value=Change.FORMED)
        assert (
            crg.get_bond_attribute(0, 1, attr="reaction") == Change.FORMED
        )

    def test_set_bond_attribute(self, crg):
        crg.set_bond_attribute(0, 1, attr="bond_label", value="test")
        assert crg.get_bond_attribute(0, 1, attr="bond_label") == "test"

    def test_add_formed_bond(self, crg):
        crg.add_formed_bond(1, 2)
        assert (
            crg.get_bond_attribute(1, 2, attr="reaction") == Change.FORMED
        )

    def test_add_broken_bond(self, crg):
        crg.add_broken_bond(1, 2)
        assert (
            crg.get_bond_attribute(2, 1, attr="reaction") == Change.BROKEN
        )

    def test_get_formed_bonds(self, crg):
        assert set(crg.get_formed_bonds()) == {Bond((0, 1)),}

    def test_get_broken_bonds(self, crg):
        assert set(crg.get_broken_bonds()) == {Bond((2, 3)),}

    def test_active_atoms_crg(self, crg):
        assert set(crg.active_atoms(additional_layer = 0)) == {0, 1, 2, 3}
        assert set(crg.active_atoms(additional_layer = 1)) == {0, 1, 2, 3}

        crg.add_atom(4, "O")
        crg.add_bond(4, 2)
        crg.add_atom(5, "H")
        crg.add_bond(5, 4)

        assert set(crg.active_atoms(additional_layer = 0)) == {0, 1, 2, 3}
        assert set(crg.active_atoms(additional_layer = 1)) == {0, 1, 2, 3, 4}

    def test_active_atoms(self):
        crg = self._TestClass()
        crg.add_atom(1, "C")
        crg.add_atom(0, "C")
        crg.add_broken_bond(0,1)
        for i in range(2, 10):
            crg.add_atom(i, "C")
            crg.add_bond(i, i-1)
        for j in range(0, 5):
            assert set(crg.active_atoms( additional_layer=j)) == {*range(j+2)}

    def test_reactant_with_attributes(self, crg):
        crg_copy = crg.copy()
        for bond in crg.get_formed_bonds():
            crg_copy.remove_bond(*bond)
        for bond in crg.get_broken_bonds():
            crg_copy.delete_bond_attribute(*bond, attr="reaction")

        expected_result = self.__class__.__bases__[0]._TestClass(crg_copy)
        assert (
            expected_result.bonds_with_attributes
            == crg.reactant(keep_attributes=True).bonds_with_attributes
        )

    def test_reactant_without_attributes(self, crg):
        crg_copy = crg.copy()
        for bond in crg.get_formed_bonds():
            crg_copy.remove_bond(*bond)
        for bond in crg.get_broken_bonds():
            crg_copy.delete_bond_attribute(*bond, attr="reaction")
        for bond, attr_dict in crg_copy.bonds_with_attributes.items():
            to_delete = [bond_attr for bond_attr in attr_dict.keys()]
            for bond_attr in to_delete:
                crg_copy.delete_bond_attribute(*bond, attr=bond_attr)
        for atom, attr_dict in crg_copy.atoms_with_attributes.items():
            to_delete = [
                atom_attr
                for atom_attr in attr_dict.keys()
                if atom_attr != "atom_type"
            ]
            for atom_attr in to_delete:
                crg_copy.delete_atom_attribute(atom, atom_attr)
        expected_result = self.__class__.__bases__[0]._TestClass(crg_copy)
        assert (
            expected_result.bonds_with_attributes
            == crg.reactant(keep_attributes=False).bonds_with_attributes
        )

    def test_product_with_attributes(self, crg):
        crg_copy = crg.copy()
        for bond in crg.get_broken_bonds():
            crg_copy.remove_bond(*bond)
        for bond in crg.get_formed_bonds():
            crg_copy.delete_bond_attribute(*bond, attr="reaction")

        expected_result = self.__class__.__bases__[0]._TestClass(crg_copy)
        assert (
            expected_result.bonds_with_attributes
            == crg.product(keep_attributes=True).bonds_with_attributes
        )

    def test_product_without_attributes(self, crg):
        crg_copy = crg.copy()
        for bond in crg.get_broken_bonds():
            crg_copy.remove_bond(*bond)
        for bond in crg.get_formed_bonds():
            crg_copy.delete_bond_attribute(*bond, attr="reaction")
        for bond, attr_dict in crg_copy.bonds_with_attributes.items():
            to_delete = [bond_attr for bond_attr in attr_dict.keys()]
            for bond_attr in to_delete:
                crg_copy.delete_bond_attribute(*bond, attr=bond_attr)
        for atom, attr_dict in crg_copy.atoms_with_attributes.items():
            to_delete = [
                atom_attr
                for atom_attr in attr_dict.keys()
                if atom_attr != "atom_type"
            ]
            for atom_attr in to_delete:
                crg_copy.delete_atom_attribute(atom, atom_attr)
        expected_result = self.__class__.__bases__[0]._TestClass(crg_copy)
        assert (
            expected_result.bonds_with_attributes
            == crg.product(keep_attributes=False).bonds_with_attributes
        )

    @pytest.fixture
    def chiral_reaction_scrg1(self, chiral_reactant_geo, chiral_product_geo1):
        return self._TestClass.from_reactant_and_product_geometry(
            chiral_reactant_geo, chiral_product_geo1
        )

    @pytest.fixture
    def chiral_reaction_scrg2(self, chiral_reactant_geo, chiral_product_geo2):
        return self._TestClass.from_reactant_and_product_geometry(
            chiral_reactant_geo, chiral_product_geo2
        )


    def test_reverse_reaction(self, chiral_reaction_scrg1):
        reversed_reaction = chiral_reaction_scrg1.reverse_reaction()
        assert (
            reversed_reaction.get_broken_bonds()
            == chiral_reaction_scrg1.get_formed_bonds()
        )
        assert (
            reversed_reaction.get_formed_bonds()
            == chiral_reaction_scrg1.get_broken_bonds()
        )
        
        double_reverset_reaction = reversed_reaction.reverse_reaction()
        assert double_reverset_reaction == chiral_reaction_scrg1

    def test_isomorphism_same_reactant_and_product_without_ts(
        self, chiral_reaction_scrg1, chiral_reaction_scrg2
    ):
        assert chiral_reaction_scrg1.product().is_isomorphic(
            chiral_reaction_scrg2.product()
        )
        assert chiral_reaction_scrg1.reactant().is_isomorphic(
            chiral_reaction_scrg2.reactant()
        )
        assert chiral_reaction_scrg1.is_isomorphic(chiral_reaction_scrg2)


class TestPlanar:
    def test_are_planar_true(self):
        coords = np.array([[0.0, 0.0, 0.0],[1.0, 0.0, 0.0],[0.0, 1.0, 0.0],[1.0, 1.0, 0.1]], dtype=np.float64)
        assert are_planar(coords, threshold=0.5)

    def test_are_planar_false(self):
        coords = np.array([[0.0, 0.0, 0.0],[1.0, 0.0, 0.0],[0.0, 1.0, 0.0],[1.0, 1.0, 1.0]], dtype=np.float64)
        assert not are_planar(coords, threshold=0.5)


class TestTetrahedral:
    def test_from_coords(self, enantiomer_geos):
        coords1 = enantiomer_geos[0].coords
        coords2 = enantiomer_geos[1].coords
        atoms = (3, 0, 1, 2, 4)

        stereo1 = Tetrahedral.from_coords(atoms, coords1.take(atoms, axis=0))
        stereo2 = Tetrahedral.from_coords(atoms, coords2.take(atoms, axis=0))
        assert stereo1.parity == 1
        assert stereo2.parity == -1

    def test_from_permuted_coords(self, enantiomer_geos):
        coords = enantiomer_geos[0].coords
        different_perms = {Tetrahedral.from_coords(atoms:=(3, *perm),
                                                   coords.take(atoms, axis=0))
                    for perm in permutations((0, 1, 2, 4))}
        assert len(different_perms) == 1

    def test_equality(self):
        stereo1 = Tetrahedral((6 ,0, 1, 2, 3), 1)
        stereo2 = Tetrahedral((6, 1, 2, 0, 3), 1)
        stereo3 = Tetrahedral((6, 0, 2, 1, 3), -1)
        stereo4 = Tetrahedral((6, 0, 2, 1, 3), 1)
        assert stereo1 == stereo2 == stereo3 != stereo4
        assert hash(stereo1) == hash(stereo2) == hash(stereo3) != hash(stereo4)

    def test_equality_with_none(self):
        stereo1 = Tetrahedral((6, 0, 1, 2, 3), None)
        stereo2 = Tetrahedral((6, 1, 2, 3, 0), None)
        assert stereo1 == stereo2

    def test_permutations(self):
        stereo1 = Tetrahedral((6, 0, 1, 2, 3), 1)
        stereo2 = Tetrahedral((6, 1, 2, 0, 3), 1)
        assert set(stereo1._perm_atoms()) == set(stereo2._perm_atoms())


class TestTrigonalBipyramidal:
    def test_from_coords(self, data_path):
        pcl5 = Geometry.from_xyz_file(data_path / "PCl5.xyz")
        result = TrigonalBipyramidal.from_coords(
            (0, 1, 2, 3, 4, 5), pcl5.coords.take((0, 1, 2, 3, 4, 5), axis=0)
        )
        assert result.parity == 1
        assert set(result.atoms) == {3, 4, 0, 1, 2, 5}

    def test_equality(self):
        stereo1 = TrigonalBipyramidal(
            (6, 0, 1, 2, 3, 4), 1
        )
        stereo2 = TrigonalBipyramidal(
            (6, 1, 0, 2, 3, 4), -1
        )
        stereo3 = TrigonalBipyramidal(
            (6, 1, 0, 3, 4, 2), -1
        )
        stereo4 = TrigonalBipyramidal(
            (6, 1, 0, 3, 4, 2), 1
        )
        stereo5 = deepcopy(stereo4)
        assert stereo1 == stereo2 == stereo3 != stereo4
        assert (
            hash(stereo1)
            == hash(stereo2)
            == hash(stereo3)
            != hash(stereo4)
            == hash(stereo5)
        )

    def test_equality_with_none(self):
        stereo1 = TrigonalBipyramidal(
            (0, 1, 2, 3, 4, 5), None
        )
        stereo2 = TrigonalBipyramidal(
            (1, 0, 2, 3, 4, 5), None
        )
        assert stereo1 == stereo2


class TestPlanarBond:
    def test_equality(self):
        stereo1 = PlanarBond((5, 4, 3, 2, 1, 0), 0)
        stereo2 = PlanarBond((4, 5, 3, 2, 0, 1), 0)
        stereo3 = PlanarBond((1, 0, 2, 3, 5, 4), 0)
        stereo4 = PlanarBond((4, 5, 3, 2, 1, 0), 0)
        stereo5 = deepcopy(stereo4)
        assert stereo1 == stereo2 == stereo3 != stereo4 == stereo5
        assert (
            hash(stereo1)
            == hash(stereo2)
            == hash(stereo3)
            != hash(stereo4)
            == hash(stereo5)
        )

    def test_equality_with_none(self):
        stereo3 = PlanarBond((1, 0, 2, 3, 5, 4), None)
        stereo4 = PlanarBond((5, 4, 3, 2, 1, 0), None)
        assert stereo3 == stereo4


class TestStereoMolGraph(TestMolGraph):
    _TestClass: type[StereoMolGraph] = StereoMolGraph

    def test_from_geometries1(self, chiral_reactant_graph):
        graph = chiral_reactant_graph
        expected_atom_stereo = {
            1: Tetrahedral((1, 0, 2, 3, 9), 1),
            0: Tetrahedral((0, 1, 2, 13, 14), -1),
            9: Tetrahedral((9, 1, 10, 11, 12), -1),
            5: Tetrahedral((5, 2, 6, 7, 8), -1),
        }
        expected_atom_types = [
            PTOE[atom]
            for atom in (
                "C",
                "C",
                "C",
                "H",
                "H",
                "C",
                "H",
                "H",
                "H",
                "C",
                "H",
                "H",
                "H",
                "Cl",
                "Cl",
            )
        ]
        expected_bonds = {Bond(pair) for pair in
            [(0, 1),
            (0, 2),
            (0, 13),
            (0, 14),
            (1, 2),
            (1, 3),
            (1, 9),
            (2, 4),
            (2, 5),
            (5, 6),
            (5, 7),
            (5, 8),
            (9, 10),
            (9, 11),
            (9, 12),]
        }
        assert graph.atom_types == tuple(expected_atom_types)
        assert set(graph.bonds) == expected_bonds
        assert all(
            graph.get_atom_stereo(key) == value
            for key, value in expected_atom_stereo.items()
        )

    def test_atom_stereo(self, chiral_product_graph1):
        expected = {
            1: Tetrahedral((1, 0, 3, 9, 13), 1),
            5: Tetrahedral((5, 2, 6, 7, 8), -1),
            9: Tetrahedral((9, 1, 10, 11, 12), -1),
        }
        expected2 = {
            Bond((0, 2)): PlanarBond(
                (1, 14, 0, 2, 4, 5), 0
            )
        }
        assert all(
            key in chiral_product_graph1._atom_stereo
            for key in set(expected.keys())
        )
        assert all(
            key in expected
            for key in set(chiral_product_graph1._atom_stereo.keys())
        )
        assert all(
            expected[key] == value
            for key, value in chiral_product_graph1._atom_stereo.items()
        )
        assert expected2 == chiral_product_graph1._bond_stereo

    def test_to_rdmol_double_bond(self):
        g1 = self._TestClass()
        g1.add_atom(0, atom_type="F")
        g1.add_atom(1, atom_type="H")
        g1.add_atom(2, atom_type="C")
        g1.add_atom(3, atom_type="C")
        g1.add_atom(4, atom_type="F")
        g1.add_atom(5, atom_type="H")

        g1.add_bond(0, 2)
        g1.add_bond(1, 2)
        g1.add_bond(2, 3)
        g1.add_bond(3, 4)
        g1.add_bond(3, 5)

        g2 = g1.copy()
        g3 = g1.copy()
        g1.set_bond_stereo(PlanarBond((0, 1, 2, 3, 4, 5),
                                                 0))
        g2.set_bond_stereo(PlanarBond((1, 0, 2, 3, 4, 5),
                                                    0))
        g3.set_bond_stereo(PlanarBond((0, 1, 2, 3, 4, 5),
                                                    None))
        rdmol_g1, idx_atom_map_dict_g1 = g1._to_rdmol()
        rdmol_g2, idx_atom_map_dict_g2 = g2._to_rdmol()
        rdmol_g3, idx_atom_map_dict_g3 = g3._to_rdmol()

        db1 = rdmol_g1.GetBondBetweenAtoms(2,3)
        stereo_atoms1 = {idx_atom_map_dict_g1[i] for i in db1.GetStereoAtoms()}
        assert stereo_atoms1 == {0, 4} or stereo_atoms1 == {1, 5}
        assert db1.GetStereo() == rdkit.Chem.rdchem.BondStereo.STEREOZ # type: ignore

        db2 = rdmol_g2.GetBondBetweenAtoms(2,3)
        stereo_atoms2 = {idx_atom_map_dict_g2[i] for i in db2.GetStereoAtoms()}
        assert stereo_atoms2 == {1, 4} or stereo_atoms2 == {0, 5}
        assert db2.GetStereo() == rdkit.Chem.rdchem.BondStereo.STEREOZ # type: ignore

        db3 = rdmol_g3.GetBondBetweenAtoms(2,3)

        assert db3.GetStereo() == rdkit.Chem.rdchem.BondStereo.STEREONONE # type: ignore

    def test_to_rdmol_tetrahedral(self):
        g = self._TestClass()
        g.add_atom(0, atom_type="C")
        g.add_atom(1, atom_type="H")
        g.add_atom(2, atom_type="F")
        g.add_atom(3, atom_type="Cl")
        g.add_atom(4, atom_type="Br")
        g.add_bond(0, 1)
        g.add_bond(0, 2)
        g.add_bond(0, 3)
        g.add_bond(0, 4)
        g.set_atom_stereo(Tetrahedral((0, 1, 2, 3, 4), 1))

        mol, _ = g._to_rdmol()
        chiral_tag = rdkit.Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW # type: ignore
        assert mol.GetAtomWithIdx(0).GetChiralTag() == chiral_tag

        g.set_atom_stereo(Tetrahedral((0, 1, 2, 3, 4), -1))
        mol, _ = g._to_rdmol()
        chiral_tag = rdkit.Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW # type: ignore
        assert mol.GetAtomWithIdx(0).GetChiralTag() == chiral_tag

    @pytest.mark.parametrize("inchi", [
        (r"InChI=1S/CHBrClF/c2-1(3)4/h1H/t1-/m0/s1"),
        (r"InChI=1S/C2H2Cl2/c3-1-2-4/h1-2H/b2-1+"),
        (r"InChI=1S/C2H2Cl2/c3-1-2-4/h1-2H/b2-1-"),
        (r"InChI=1S/C4H6/c1-3-4-2/h3-4H,1-2H2")
    ], ids = ["(R)-Bromochlorofluoromethane",
              "Trans-1,2-Dichloroethylene",
              "Cis-1,2-Dichloroethylene",
              "Butadiene"],)
    def test_from_rdmol_to_rdmol_stereo(self, inchi):
        rdmol = rdkit.Chem.MolFromInchi(inchi, sanitize=False)
        rdmol = rdkit.Chem.AddHs(rdmol, explicitOnly=True)
        molgraph = self._TestClass.from_rdmol(rdmol)
        rdmol2, _ = molgraph._to_rdmol(generate_bond_orders=True)
        assert inchi == rdkit.Chem.MolToInchi(rdmol2, treatWarningAsError=True) # type: ignore

    @pytest.mark.parametrize("smiles", [
        "[H][Pt@SP1](F)(Cl)Br",
        "Cl[Pt@SP1](Cl)([NH3])[NH3]",
        "Cl[Pt@SP2](Cl)([NH3])[NH3]"
    ], ids = ["Hydridofluorochlorobromoplatinum(II)",
              "(SP-4-2)-diamminedichloroplatinum",
              "(SP-4-1)-diamminedichloroplatinum"],)
    def test_from_rdmol_to_rdmol_square_planar(self, smiles):
        rdmol = rdkit.Chem.MolFromSmiles(smiles, sanitize=True)
        rdmol = rdkit.Chem.AddHs(rdmol, explicitOnly=True)
        molgraph = self._TestClass.from_rdmol(rdmol)
        #raise Exception (molgraph.stereo)
        rdmol2, _ = molgraph._to_rdmol(generate_bond_orders=True,
                                       allow_charged_fragments=True)
        rdkit.Chem.SanitizeMol(rdmol2, sanitizeOps=rdkit.Chem.SanitizeFlags.SANITIZE_ALL)
        for atom in rdmol2.GetAtoms():
            atom.SetAtomMapNum(0)
        assert rdkit.Chem.MolToSmiles(rdmol) == rdkit.Chem.MolToSmiles(rdmol2)

    def test_from_rdmol_square_planar_different(self):
        smiles = ["Cl[Pt@SP1](Cl)([NH3])[NH3]", "Cl[Pt@SP2](Cl)([NH3])[NH3]"]
        rdmols = [rdkit.Chem.MolFromSmiles(s, sanitize=False) for s in smiles]
        rdmol = [rdkit.Chem.AddHs(mol, explicitOnly=True) for mol in rdmols]
        molgraphs = [self._TestClass.from_rdmol(mol) for mol in rdmol]
        assert molgraphs[0] != molgraphs[1]

    def test_from_rdmol_square_planar(self):
        smiles = ("C[Pt@SP1](F)(Cl)[H]", "C[Pt@SP2](Cl)(F)[H]",
                  "C[Pt@SP3](F)([H])Cl")
        mols = [rdkit.Chem.MolFromSmiles(i, sanitize=False) for i in smiles]
        mols = [rdkit.Chem.AddHs(i, explicitOnly=True) for i in mols]
        molgraphs = [self._TestClass.from_rdmol(i) for i in mols]
        assert all(molgraph == molgraphs[0] for molgraph in molgraphs)

    def test_from_rdmol_trigonal_bipyramidal(self):
        smiles = ("S[As@TB1](F)(Cl)(Br)N", "S[As@TB2](F)(Br)(Cl)N",
                  "S[As@TB3](F)(Cl)(N)Br", "S[As@TB4](F)(Br)(N)Cl",
                  "S[As@TB5](F)(N)(Cl)Br", "S[As@TB6](F)(N)(Br)Cl",
                  "S[As@TB7](N)(F)(Cl)Br", "S[As@TB8](N)(F)(Br)Cl",
                  "F[As@TB9](S)(Cl)(Br)N", "F[As@TB11](S)(Br)(Cl)N",
                  "F[As@TB10](S)(Cl)(N)Br", "F[As@TB12](S)(Br)(N)Cl",
                  "F[As@TB13](S)(N)(Cl)Br", "F[As@TB14](S)(N)(Br)Cl",
                  "F[As@TB15](Cl)(S)(Br)N", "F[As@TB20](Br)(S)(Cl)N",
                  "F[As@TB16](Cl)(S)(N)Br", "F[As@TB19](Br)(S)(N)Cl",
                  "F[As@TB17](Cl)(Br)(S)N", "F[As@TB18](Br)(Cl)(S)N")
        mols = [rdkit.Chem.MolFromSmiles(i, sanitize=False) for i in smiles]
        mols = [rdkit.Chem.AddHs(i, explicitOnly=True) for i in mols]
        molgraphs = [self._TestClass.from_rdmol(i) for i in mols]
        assert all(any( isinstance(stereo, TrigonalBipyramidal) for
                       stereo in molgraph.stereo.values())
                   for molgraph in molgraphs)
        assert all(molgraph == molgraphs[0] for molgraph in molgraphs)


    def test_from_rdmol_octahedral(self):
        smiles = ("O[Co@OH1](Cl)(C)(N)(F)P", "O[Co@OH2](Cl)(F)(N)(C)P",
                  "O[Co@OH3](Cl)(C)(N)(P)F", "O[Co@OH16](Cl)(F)(N)(P)C",
                  "O[Co@OH6](Cl)(C)(P)(N)F", "O[Co@OH18](Cl)(F)(P)(N)C",
                  "O[Co@OH19](Cl)(P)(C)(N)F", "O[Co@OH24](Cl)(P)(F)(N)C",
                  "O[Co@OH25](P)(Cl)(C)(N)F", "O[Co@OH30](P)(Cl)(F)(N)C",
                  "O[Co@OH4](Cl)(C)(F)(N)P", "O[Co@OH14](Cl)(F)(C)(N)P",
                  "O[Co@OH5](Cl)(C)(F)(P)N", "O[Co@OH15](Cl)(F)(C)(P)N",
                  "O[Co@OH7](Cl)(C)(P)(F)N", "O[Co@OH17](Cl)(F)(P)(C)N",
                  "O[Co@OH20](Cl)(P)(C)(F)N", "O[Co@OH23](Cl)(P)(F)(C)N",
                  "O[Co@OH26](P)(Cl)(C)(F)N", "O[Co@OH29](P)(Cl)(F)(C)N",
                  "O[Co@OH10](Cl)(N)(F)(C)P", "O[Co@OH8](Cl)(N)(C)(F)P",
                  "O[Co@OH11](Cl)(N)(F)(P)C", "O[Co@OH9](Cl)(N)(C)(P)F",
                  "O[Co@OH13](Cl)(N)(P)(F)C", "O[Co@OH12](Cl)(N)(P)(C)F",
                  "O[Co@OH22](Cl)(P)(N)(F)C", "O[Co@OH21](Cl)(P)(N)(C)F",
                  "O[Co@OH28](P)(Cl)(N)(F)C", "O[Co@OH27](P)(Cl)(N)(C)F")

        mols = [rdkit.Chem.MolFromSmiles(i, sanitize=False) for i in smiles]
        mols = [rdkit.Chem.AddHs(i, explicitOnly=True) for i in mols]
        molgraphs = [self._TestClass.from_rdmol(i) for i in mols]
        assert all(any( isinstance(stereo, Octahedral) for
                       stereo in molgraph.stereo.values())
                   for molgraph in molgraphs)
        assert all(molgraph == molgraphs[0] for molgraph in molgraphs)

    def test_from_rdmol_octahedral_compare(self):
        identical1 = ("Cl[Co@OH1](N)(N)(O)(Cl)Cl", "Cl[Co@OH2](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH3](N)(N)(O)(Cl)Cl", "Cl[Co@OH4](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH5](N)(N)(O)(Cl)Cl","Cl[Co@OH14](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH15](N)(N)(O)(Cl)Cl","Cl[Co@OH16](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH21](N)(N)(O)(Cl)Cl","Cl[Co@OH22](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH27](N)(N)(O)(Cl)Cl","Cl[Co@OH28](N)(N)(O)(Cl)Cl")
        identical2 = ("Cl[Co@OH6](N)(N)(O)(Cl)Cl","Cl[Co@OH7](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH17](N)(N)(O)(Cl)Cl","Cl[Co@OH18](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH19](N)(N)(O)(Cl)Cl","Cl[Co@OH20](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH23](N)(N)(O)(Cl)Cl","Cl[Co@OH24](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH25](N)(N)(O)(Cl)Cl","Cl[Co@OH26](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH29](N)(N)(O)(Cl)Cl","Cl[Co@OH30](N)(N)(O)(Cl)Cl")
        identical3 = ("Cl[Co@OH8](N)(N)(O)(Cl)Cl","Cl[Co@OH9](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH10](N)(N)(O)(Cl)Cl","Cl[Co@OH11](N)(N)(O)(Cl)Cl",
                      "Cl[Co@OH12](N)(N)(O)(Cl)Cl","Cl[Co@OH13](N)(N)(O)(Cl)Cl")

        mols1 = [rdkit.Chem.MolFromSmiles(i, sanitize=False)
                 for i in identical1]
        mols1 = [rdkit.Chem.AddHs(i, explicitOnly=True) for i in mols1]
        molgraphs1 = [self._TestClass.from_rdmol(i) for i in mols1]
        mols2 = [rdkit.Chem.MolFromSmiles(i, sanitize=False)
                 for i in identical2]
        mols2 = [rdkit.Chem.AddHs(i, explicitOnly=True) for i in mols2]
        molgraphs2 = [self._TestClass.from_rdmol(i) for i in mols2]
        mols3 = [rdkit.Chem.MolFromSmiles(i, sanitize=False)
                 for i in identical3]
        mols3 = [rdkit.Chem.AddHs(i, explicitOnly=True) for i in mols3]
        molgraphs3 = [self._TestClass.from_rdmol(i) for i in mols3]

        assert all(molgraph == molgraphs1[0] for molgraph in molgraphs1)
        assert all(molgraph == molgraphs2[0] for molgraph in molgraphs2)
        assert all(molgraph == molgraphs3[0] for molgraph in molgraphs3)
        assert molgraphs1[0] != molgraphs2[0]
        assert molgraphs1[0] != molgraphs3[0]
        assert molgraphs2[0] != molgraphs3[0]

        set1 = {rdkit.Chem.MolToSmiles(molgraph._to_rdmol()[0],
                                       canonical=True,
                                       ignoreAtomMapNumbers=True,
                                       isomericSmiles=True)
                   for molgraph in molgraphs1}
        set2 = {rdkit.Chem.MolToSmiles(molgraph._to_rdmol()[0],
                                       canonical=True,
                                       ignoreAtomMapNumbers=True,
                                       isomericSmiles=True)
                   for molgraph in molgraphs2}
        set3 = {rdkit.Chem.MolToSmiles(molgraph._to_rdmol()[0],
                                       canonical=True,
                                       ignoreAtomMapNumbers=True,
                                       isomericSmiles=True)
                   for molgraph in molgraphs3}

        assert set() == set1 & set2 == set1 & set3 == set2 & set3

    def test_from_atrop(self):
        self._TestClass()
        smg = StereoMolGraph()
        smg.add_atom(0, "H")
        smg.add_atom(1, "Cl")
        smg.add_atom(2, "C")
        smg.add_atom(3, "C")
        smg.add_atom(4, "F")
        smg.add_atom(5, "I")

        smg.add_bond(0,2)
        smg.add_bond(1,2)
        smg.add_bond(2,3)
        smg.add_bond(3,4)
        smg.add_bond(3,5)

        smg.set_bond_stereo(AtropBond(atoms=(0, 1, 2, 3, 4, 5), parity=1))

    def test_from_composed_chiral_molgraphs(
        self, chiral_product_graph1, chiral_product_graph2
    ):
        relabel_mapping = {
            atom: atom + chiral_product_graph1.n_atoms
            for atom in chiral_product_graph2.atoms
        }
        chiral_product_graph2.relabel_atoms(relabel_mapping, copy=False)

        combined = self._TestClass.compose(
            [chiral_product_graph1, chiral_product_graph2]
        )

        assert (
            combined.atoms_with_attributes
            == chiral_product_graph1.atoms_with_attributes
            | chiral_product_graph2.atoms_with_attributes
        )
        assert (
            combined.bonds_with_attributes
            == chiral_product_graph1.bonds_with_attributes
            | chiral_product_graph2.bonds_with_attributes
        )
        assert (
            combined.stereo
            == chiral_product_graph1.stereo
            | chiral_product_graph2.stereo
        )

    def test_atom_stereo_is_isomorphic(self, chiral_product_graph1):
        isomorphic_graph = chiral_product_graph1.copy()
        isomorphic_graph.relabel_atoms(
            {0: 1, 1: 0, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7}
        )
        assert chiral_product_graph1.is_isomorphic(isomorphic_graph)

    def test_get_atom_stereo_isomorphic_mappings(self, chiral_product_graph1):
        isomorphic_graph = chiral_product_graph1.copy()
        isomorphic_graph.relabel_atoms(
            {0: 1, 1: 0, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7}
        )
        assert {
            frozenset(mapping.items()) for mapping in
            chiral_product_graph1.get_isomorphic_mappings(isomorphic_graph)
         } == {frozenset(mapping.items()) for mapping in (
            {
                0: 0,
                1: 1,
                2: 2,
                3: 3,
                4: 4,
                5: 5,
                6: 6,
                7: 7,
                8: 8,
                9: 9,
                10: 10,
                11: 11,
                12: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                1: 1,
                2: 2,
                3: 3,
                4: 4,
                5: 5,
                6: 6,
                7: 7,
                8: 8,
                9: 9,
                11: 10,
                12: 11,
                10: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                1: 1,
                2: 2,
                3: 3,
                4: 4,
                5: 5,
                6: 6,
                7: 7,
                8: 8,
                9: 9,
                12: 10,
                10: 11,
                11: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                1: 1,
                2: 2,
                3: 3,
                4: 4,
                5: 5,
                7: 6,
                8: 7,
                6: 8,
                9: 9,
                10: 10,
                11: 11,
                12: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                1: 1,
                2: 2,
                3: 3,
                4: 4,
                5: 5,
                7: 6,
                8: 7,
                6: 8,
                9: 9,
                11: 10,
                12: 11,
                10: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                1: 1,
                2: 2,
                3: 3,
                4: 4,
                5: 5,
                7: 6,
                8: 7,
                6: 8,
                9: 9,
                12: 10,
                10: 11,
                11: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                1: 1,
                2: 2,
                3: 3,
                4: 4,
                5: 5,
                8: 6,
                6: 7,
                7: 8,
                9: 9,
                10: 10,
                11: 11,
                12: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                1: 1,
                2: 2,
                3: 3,
                4: 4,
                5: 5,
                8: 6,
                6: 7,
                7: 8,
                9: 9,
                11: 10,
                12: 11,
                10: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                1: 1,
                2: 2,
                3: 3,
                4: 4,
                5: 5,
                8: 6,
                6: 7,
                7: 8,
                9: 9,
                12: 10,
                10: 11,
                11: 12,
                13: 13,
                14: 14,
            },
        )}

    def test_get_isomorphic_mappings_of_enantiomers(
        self, enantiomer_graph1, enantiomer_graph2
    ):
        assert not enantiomer_graph1.is_isomorphic(enantiomer_graph2)

    def test_get_isomorphic_mappings(
        self, chiral_product_graph1, chiral_product_graph2
    ):
        expected_isomorphisms = (
            {
                0: 0,
                2: 1,
                1: 2,
                4: 3,
                3: 4,
                9: 5,
                10: 6,
                11: 7,
                12: 8,
                5: 9,
                6: 10,
                7: 11,
                8: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                2: 1,
                1: 2,
                4: 3,
                3: 4,
                9: 5,
                10: 6,
                11: 7,
                12: 8,
                5: 9,
                7: 10,
                8: 11,
                6: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                2: 1,
                1: 2,
                4: 3,
                3: 4,
                9: 5,
                10: 6,
                11: 7,
                12: 8,
                5: 9,
                8: 10,
                6: 11,
                7: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                2: 1,
                1: 2,
                4: 3,
                3: 4,
                9: 5,
                11: 6,
                12: 7,
                10: 8,
                5: 9,
                6: 10,
                7: 11,
                8: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                2: 1,
                1: 2,
                4: 3,
                3: 4,
                9: 5,
                11: 6,
                12: 7,
                10: 8,
                5: 9,
                7: 10,
                8: 11,
                6: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                2: 1,
                1: 2,
                4: 3,
                3: 4,
                9: 5,
                11: 6,
                12: 7,
                10: 8,
                5: 9,
                8: 10,
                6: 11,
                7: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                2: 1,
                1: 2,
                4: 3,
                3: 4,
                9: 5,
                12: 6,
                10: 7,
                11: 8,
                5: 9,
                6: 10,
                7: 11,
                8: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                2: 1,
                1: 2,
                4: 3,
                3: 4,
                9: 5,
                12: 6,
                10: 7,
                11: 8,
                5: 9,
                7: 10,
                8: 11,
                6: 12,
                13: 13,
                14: 14,
            },
            {
                0: 0,
                2: 1,
                1: 2,
                4: 3,
                3: 4,
                9: 5,
                12: 6,
                10: 7,
                11: 8,
                5: 9,
                8: 10,
                6: 11,
                7: 12,
                13: 13,
                14: 14,
            },
        )

        assert chiral_product_graph1.is_isomorphic(chiral_product_graph2)
        assert chiral_product_graph1.is_isomorphic(chiral_product_graph2)
        assert all(
            i in expected_isomorphisms
            for i in chiral_product_graph1.get_isomorphic_mappings(
                chiral_product_graph2
            )
        )

    def test_enantiomer(self, enantiomer_graph1, enantiomer_graph2):
        assert enantiomer_graph1.enantiomer().is_isomorphic(enantiomer_graph2)
        assert not enantiomer_graph1.is_isomorphic(enantiomer_graph2)

    def test_hash_enantiomers(self, enantiomer_graph1, enantiomer_graph2):
        assert enantiomer_graph1._atom_stereo != enantiomer_graph2._atom_stereo
        #raise Exception(enantiomer_graph1._atom_stereo, enantiomer_graph2._atom_stereo)
        assert (hash(enantiomer_graph1)
                != hash(enantiomer_graph2))

    def test_valid_stereo(self, chiral_product_graph1):
        assert chiral_product_graph1.is_stereo_valid()


class TestStereoCondensedReactionGraph(
    TestStereoMolGraph, TestCondensedReactionGraph
):
    _TestClass: type[StereoCondensedReactionGraph] = (
        StereoCondensedReactionGraph
    )

    @pytest.fixture
    def chiral_ts_geo1(self, data_path):
        return Geometry.from_xyz_file(
            data_path/ "conrot_reaction/ts.xyz")

    @pytest.fixture
    def chiral_ts_geo2(self, data_path):
        return Geometry.from_xyz_file(
            data_path / "disrot_reaction/ts.xyz")
        

    def test_product_with_attributes(self, crg):
        super().test_product_with_attributes(crg)
        expected_mol_atom_stereo = {
            key: value
            for key, value in crg._atom_stereo.items()
            if value in (None, Change.FORMED)
        }
        assert (
            crg.product(keep_attributes=True)._atom_stereo
            == expected_mol_atom_stereo
        )

    def test_product_without_attributes(self, crg):
        super().test_product_without_attributes(crg)
        expected_mol_atom_stereo = {
            key: value
            for key, value in crg._atom_stereo.items()
            if value in (None, Change.FORMED)
        }
        assert (
            crg.product(keep_attributes=True)._atom_stereo
            == expected_mol_atom_stereo
        )

    def test_reactant_with_attributes(self, crg):
        super().test_reactant_with_attributes(crg)
        expected_mol_atom_stereo = {
            key: value
            for key, value in crg._atom_stereo.items()
            if value in (None, Change.BROKEN)
        }
        assert (
            crg.product(keep_attributes=True)._atom_stereo
            == expected_mol_atom_stereo
        )

    def test_reactant_without_attributes(self, crg):
        super().test_reactant_without_attributes(crg)
        expected_mol_atom_stereo = {
            key: value
            for key, value in crg._atom_stereo.items()
            if value in (None, Change.BROKEN)
        }
        assert (
            crg.product(keep_attributes=True)._atom_stereo
            == expected_mol_atom_stereo
        )

    def test_from_composed_chiral_molgraphs(
        self, chiral_product_graph1, chiral_product_graph2
    ):
        relabel_mapping = {
            atom: atom + chiral_product_graph1.n_atoms
            for atom in chiral_product_graph2.atoms
        }
        chiral_product_graph2.relabel_atoms(relabel_mapping, copy=False)

        combined = self._TestClass.compose(
            [chiral_product_graph1, chiral_product_graph2]
        )

        assert (
            combined.atoms_with_attributes
            == chiral_product_graph1.atoms_with_attributes
            | chiral_product_graph2.atoms_with_attributes
        )
        assert (
            combined.bonds_with_attributes
            == chiral_product_graph1.bonds_with_attributes
            | chiral_product_graph2.bonds_with_attributes
        )
        assert (
            combined.stereo
            == chiral_product_graph1.stereo
            | chiral_product_graph2.stereo
        )
        assert (
            {**combined.atom_stereo_changes, **combined.bond_stereo_changes}
            == {**chiral_product_graph1.atom_stereo_changes,
                **chiral_product_graph1.bond_stereo_changes,
                **chiral_product_graph2.atom_stereo_changes,
                **chiral_product_graph2.bond_stereo_changes}
        )

    def test_from_chain_of_states_reaction(self, data_path):
        reactant_geo = Geometry.from_xyz_file(
            data_path / "methylamine_phosgenation_trans_r.xyz"
            
        )
        product_geo = Geometry.from_xyz_file(
            data_path / "methylamine_phosgenation_trans_p.xyz"
        )
        ts_geo =  Geometry.from_xyz_file(
            data_path / "methylamine_phosgenation_trans_ts.xyz"
        )
        
        scrg = self._TestClass.from_reactant_product_and_ts_geometry(
            reactant_geo, product_geo, ts_geo
        )

        atoms = {
            0: {"atom_type": PTOE["H"]},
            1: {"atom_type": PTOE["N"]},
            2: {"atom_type": PTOE["H"]},
            3: {"atom_type": PTOE["C"]},
            4: {"atom_type": PTOE["Cl"]},
            5: {"atom_type": PTOE["O"]},
            6: {"atom_type": PTOE["Cl"]},
            7: {"atom_type": PTOE["C"]},
            8: {"atom_type": PTOE["H"]},
            9: {"atom_type": PTOE["H"]},
            10: {"atom_type": PTOE["H"]},
        }
        bonds = {
            Bond((0, 1)): {},
            Bond((1, 2)): {"reaction": Change.BROKEN},
            Bond((1, 7)): {},
            Bond((1, 3)): {},
            Bond((2, 6)): {"reaction": Change.FORMED},
            Bond((3, 4)): {},
            Bond((3, 6)): {"reaction": Change.BROKEN},
            Bond((3, 5)): {},
            Bond((7, 10)): {},
            Bond((7, 9)): {},
            Bond((7, 8)): {},
        }
        stereo = {7: Tetrahedral((7, 1, 8, 9, 10), -1)}
        atom_stereo_change = defaultdict(ChangeDict,{
            1: ChangeDict({
                Change.BROKEN: Tetrahedral(
                    (1, 0, 2, 3, 7), -1
                )
            }),
            3: ChangeDict({
                Change.BROKEN: Tetrahedral(
                    (3, 1, 4, 5, 6), -1
                )
            }),})
        bond_stereo_change = defaultdict(ChangeDict,{
            Bond({1, 3}): ChangeDict({
                Change.FORMED: PlanarBond(
                    (4, 5, 3, 1, 0, 7), 0
                )})})
        assert scrg.atoms_with_attributes == atoms
        assert scrg.bonds_with_attributes == bonds
        assert scrg.stereo == stereo
        assert scrg._atom_stereo_change == atom_stereo_change
        assert scrg._bond_stereo_change == bond_stereo_change

    @pytest.fixture
    def scrg_stereo_inversion(self, data_path):
        """ A SCRG with just the inversion of a stereocenter, reactant and product are tetrahedral but
        the transition state is square planar. """
        reactant_geo = Geometry.from_xyz_file(data_path / "fluoro_chloro_bromomethane_r.xyz")
        product_geo = Geometry.from_xyz_file(data_path / "fluoro_chloro_bromomethane_s.xyz")
        ts_geo =  Geometry.from_xyz_file(data_path / "fluoro_chloro_bromomethane_ts.xyz")
        
        scrg = self._TestClass.from_reactant_product_and_ts_geometry(
            reactant_geo, product_geo, ts_geo
        )
        return scrg

    def test_creation_from_xyz_atom_stereocenter_inversion(
        self, scrg_stereo_inversion
    ):
        scrg = scrg_stereo_inversion
        assert scrg.get_atom_stereo_change(0) == {
            Change.BROKEN: Tetrahedral(
                (0, 1, 2, 3, 4), 1
            ),
            Change.FORMED: Tetrahedral(
                (0, 1, 2, 3, 4), -1
            ),
        }
        assert scrg._bond_stereo_change == {}

    def test_relabel_reaction_atoms(self, scrg_stereo_inversion):
        scrg = scrg_stereo_inversion
        scrg.relabel_atoms({0: 11, 1: 10, 2: 20, 3: 30, 4: 40}, copy=False)
        assert scrg.get_atom_stereo_change(11) == {
            Change.BROKEN: Tetrahedral(
                (11, 10, 20, 30, 40), 1
            ),
            Change.FORMED: Tetrahedral(
                (11, 10, 20, 30, 40), -1
            ),
        }
        assert scrg._bond_stereo_change == {}

    def test_relabel_reaction_atoms_copy(self, scrg_stereo_inversion):
        scrg = scrg_stereo_inversion
        new_scrg = scrg.relabel_atoms(
            {0: 11, 1: 10, 2: 20, 3: 30, 4: 40}, copy=True
        )
        assert new_scrg.get_atom_stereo_change(11) == {
            Change.BROKEN: Tetrahedral(
                (11, 10, 20, 30, 40), 1
            ),
            Change.FORMED: Tetrahedral(
                (11, 10, 20, 30, 40), -1
            ),
        }
        assert scrg._bond_stereo_change == {}

    @pytest.fixture
    def chiral_reaction_chiral_ts_scrg1(
        self, chiral_reactant_geo, chiral_product_geo1, chiral_ts_geo1
    ):
        return self._TestClass.from_reactant_product_and_ts_geometry(
            chiral_reactant_geo, chiral_product_geo1, chiral_ts_geo1
        )

    @pytest.fixture
    def chiral_reaction_chiral_ts_scrg2(
        self, chiral_reactant_geo, chiral_product_geo2, chiral_ts_geo2
    ):
        return self._TestClass.from_reactant_product_and_ts_geometry(
            chiral_reactant_geo, chiral_product_geo2, chiral_ts_geo2
        )

    def test_isomorphism_same_reactant_and_product_without_ts(
        self, chiral_reaction_scrg1, chiral_reaction_scrg2
    ):
        assert chiral_reaction_scrg1.product().is_isomorphic(
            chiral_reaction_scrg2.product()
        )
        assert chiral_reaction_scrg1.reactant().is_isomorphic(
            chiral_reaction_scrg2.reactant()
        )
        assert not chiral_reaction_scrg1.is_isomorphic(chiral_reaction_scrg2)

    def test_isomorphism_same_reactant_and_product_but_different_ts(
        self, chiral_reaction_chiral_ts_scrg1, chiral_reaction_chiral_ts_scrg2
    ):
        assert chiral_reaction_chiral_ts_scrg1.reactant().is_isomorphic(
            chiral_reaction_chiral_ts_scrg2.reactant()
        )
        assert chiral_reaction_chiral_ts_scrg1.product().is_isomorphic(
            chiral_reaction_chiral_ts_scrg2.product()
        )

        assert not chiral_reaction_chiral_ts_scrg1.is_isomorphic(
            chiral_reaction_chiral_ts_scrg2
        )

    def test_reverse_reaction(self, chiral_reaction_scrg1):
        reversed_reaction = chiral_reaction_scrg1.reverse_reaction()
        assert (
            reversed_reaction.get_broken_bonds()
            == chiral_reaction_scrg1.get_formed_bonds()
        )
        assert (
            reversed_reaction.get_formed_bonds()
            == chiral_reaction_scrg1.get_broken_bonds()
        )
        
        double_reverset_reaction = reversed_reaction.reverse_reaction()
        assert (chiral_reaction_scrg1.atom_stereo == double_reverset_reaction.atom_stereo)
        assert (chiral_reaction_scrg1.bond_stereo == double_reverset_reaction.bond_stereo)
        assert (chiral_reaction_scrg1.bond_stereo_changes == double_reverset_reaction.bond_stereo_changes)
        assert (chiral_reaction_scrg1.atom_stereo_changes == double_reverset_reaction.atom_stereo_changes)

        assert double_reverset_reaction == chiral_reaction_scrg1

    def test_hash_stereo_reaction(
        self, chiral_reaction_scrg1, chiral_reaction_scrg2
    ):
        assert (
            hash(chiral_reaction_scrg1)
            != hash(chiral_reaction_scrg2)
        )

    def test_hash_stereo_reaction_with_ts(
        self, chiral_reaction_chiral_ts_scrg1, chiral_reaction_chiral_ts_scrg2
    ):
        assert (
            hash(chiral_reaction_chiral_ts_scrg1)
            != hash(chiral_reaction_chiral_ts_scrg2)
        )

    def test_hash_enantiomers(
        self, enantiomer_graph1, enantiomer_graph2
    ):
        assert (hash(enantiomer_graph1)
                != hash(enantiomer_graph2))



# pyright: standard
# typing with rdkit is not fully supported
from __future__ import annotations

from typing import TYPE_CHECKING

import rdkit.Chem as Chem # type: ignore

from stereomolgraph.stereodescriptors import (
    Tetrahedral,
    SquarePlanar,
    TrigonalBipyramidal,
    Octahedral,
    PlanarBond,
    AtomStereo,
)

if TYPE_CHECKING:
    from stereomolgraph.graphs.mg import MolGraph
    from stereomolgraph.graphs.smg import StereoMolGraph


def mol_graph_from_rdmol(
    cls: type[MolGraph], rdmol: Chem.Mol, use_atom_map_number:bool=False
) -> MolGraph:
    """
    Creates a StereoMolGraph from an RDKit Mol object.
    Implicit Hydrogens are added to the graph.
    Stereo information is conserved. Double bonds, aromatic bonds and
    conjugated bonds are interpreted as planar. Atoms with 5 bonding
    partners are assumed to be TrigonalBipyramidal and allow interchange
    of the substituents (berry pseudorotation). Atoms with 6 bonding
    partners are assumed to be octahedral and do not allow interchange of
    the substituents.

    :param rdmol: RDKit Mol object
    :param use_atom_map_number: If the atom map number should be used
                                instead of the atom index
    :return: StereoMolGraph
    """
    # rdmol = Chem.AddHs(rdmol, explicitOnly=True, addCoords=True)

    if use_atom_map_number is False:
        rdmol = Chem.rdmolops.AddHs(rdmol, explicitOnly=True)

    graph = cls()

    if use_atom_map_number:
        id_atom_map = {atom.GetIdx(): atom.GetAtomMapNum() for atom in rdmol.GetAtoms()}
    else:
        id_atom_map = {atom.GetIdx(): atom.GetIdx() for atom in rdmol.GetAtoms()}

    for atom in rdmol.GetAtoms():
        graph.add_atom(id_atom_map[atom.GetIdx()], atom.GetSymbol())

    for bond in rdmol.GetBonds():
        graph.add_bond(
            id_atom_map[bond.GetBeginAtomIdx()],
            id_atom_map[bond.GetEndAtomIdx()],
        )
    return graph


def stereo_mol_graph_from_rdmol(
    cls: type[StereoMolGraph], rdmol: Chem.Mol, use_atom_map_number:bool=False
) -> StereoMolGraph:
    """
    Creates a StereoMolGraph from an RDKit Mol object.
    All hydrogens have to be explicit.
    Stereo information is conserved for tetrahedral atoms and
    double bonds.

    :param rdmol: RDKit Mol object
    :param use_atom_map_number: If the atom map number should be used
                                instead of the atom index, Default: False
    :return: StereoMolGraph
    """
    rd_tetrahedral = {
        Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW: -1,
        Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW: 1,
        Chem.rdchem.ChiralType.CHI_TETRAHEDRAL: None,
    }

    rdmol = Chem.AddHs(rdmol, explicitOnly=False)

    if use_atom_map_number is True:
        if any(atom.GetAtomMapNum() == 0 for atom in rdmol.GetAtoms()):
            raise ValueError("AtomMapNumber has to  be set on all atoms")
        id_atom_map: dict[int, int] = {
            atom.GetIdx(): atom.GetAtomMapNum() for atom in rdmol.GetAtoms()
        }
    else:
        id_atom_map: dict[int, int] = {
            atom.GetIdx(): atom.GetIdx() for atom in rdmol.GetAtoms()
        }

    graph = cls()

    for atom in rdmol.GetAtoms():
        graph.add_atom(id_atom_map[atom.GetIdx()], atom.GetSymbol())

    for bond in rdmol.GetBonds():
        graph.add_bond(
            id_atom_map[bond.GetBeginAtomIdx()],
            id_atom_map[bond.GetEndAtomIdx()],
        )

    for atom in rdmol.GetAtoms():
        atom_idx: int = atom.GetIdx()

        neighbors: tuple[int, ...] = tuple(
            [
                (
                    {b.GetBeginAtomIdx(), b.GetEndAtomIdx()}
                    - {
                        atom_idx,
                    }
                ).pop()
                for b in rdmol.GetAtomWithIdx(atom_idx).GetBonds()
            ]
        )
        neighbors: tuple[int, ...] = tuple(id_atom_map[b] for b in neighbors)
        # idx -> atom map num

        chiral_tag = atom.GetChiralTag()
        hybridization = atom.GetHybridization()
        # rad_elec = atom.GetNumRadicalElectrons()

        if len(neighbors) == 4:
            if chiral_tag in rd_tetrahedral:  #
                atom_stereo: AtomStereo = Tetrahedral(
                    (id_atom_map[atom_idx], *neighbors),
                    rd_tetrahedral[chiral_tag],
                )

                graph.set_atom_stereo(atom_stereo)

            elif hybridization == Chem.HybridizationType.SP3:
                atom_stereo = Tetrahedral((id_atom_map[atom_idx], *neighbors), None)
                graph.set_atom_stereo(atom_stereo)

        if atom.GetChiralTag() == Chem.ChiralType.CHI_SQUAREPLANAR:
            
            sp_order: tuple[int, int, int, int]
            if atom.GetUnsignedProp("_chiralPermutation") == 1:
                sp_order = (0, 1, 2, 3)
            elif atom.GetUnsignedProp("_chiralPermutation") == 2:
                sp_order = (0, 2, 1, 3)
            elif atom.GetUnsignedProp("_chiralPermutation") == 3:
                sp_order = (0, 1, 3, 2)
            else:
                raise RuntimeError("Unknown permutation for SquarePlanar")
            ordered_neighbors = tuple([neighbors[i] for i in sp_order])
            sp_atoms = (id_atom_map[atom_idx], *ordered_neighbors)
            assert len(sp_atoms) == 5
            atom_stereo = SquarePlanar(atoms=sp_atoms, parity=0)
            graph.set_atom_stereo(atom_stereo)

        if atom.GetChiralTag() == Chem.ChiralType.CHI_TRIGONALBIPYRAMIDAL:
            perm = atom.GetUnsignedProp("_chiralPermutation")

            # adapted from http://opensmiles.org/opensmiles.html
            atom_order_permutation_dict = {
                (0, 1, 2, 3, 4): 1,
                (0, 1, 3, 2, 4): 2,
                (0, 1, 2, 4, 3): 3,
                (0, 1, 4, 2, 3): 4,
                (0, 1, 3, 4, 2): 5,
                (0, 1, 4, 3, 2): 6,
                (0, 2, 3, 4, 1): 7,
                (0, 2, 4, 3, 1): 8,
                (1, 0, 2, 3, 4): 9,
                (1, 0, 3, 2, 4): 11,
                (1, 0, 2, 4, 3): 10,
                (1, 0, 4, 2, 3): 12,
                (1, 0, 3, 4, 2): 13,
                (1, 0, 4, 3, 2): 14,
                (2, 0, 1, 3, 4): 15,
                (2, 0, 1, 4, 3): 16,
                (3, 0, 1, 2, 4): 17,
                (3, 0, 2, 1, 4): 18,
                (2, 0, 4, 1, 3): 19,
                (2, 0, 3, 1, 4): 20,
            }

            permutation_atom_order_dict = {
                v: k for k, v in atom_order_permutation_dict.items()
            }

            tbp_order = permutation_atom_order_dict[perm]
            neigh_atoms = tuple([neighbors[i] for i in tbp_order])
            tbp_atoms = (id_atom_map[atom_idx], *neigh_atoms)
            assert len(tbp_atoms) == 6
            atom_stereo = TrigonalBipyramidal(tbp_atoms, 1)
            graph.set_atom_stereo(atom_stereo)

        if atom.GetChiralTag() == Chem.ChiralType.CHI_OCTAHEDRAL:
            perm = atom.GetUnsignedProp("_chiralPermutation")

            permutation_atom_order_dict = {
                1: (0, 5, 1, 2, 3, 4),
                2: (0, 5, 1, 4, 3, 2),
                3: (0, 4, 1, 2, 3, 5),
                16: (0, 4, 1, 5, 3, 2),
                6: (0, 3, 1, 2, 4, 5),
                18: (0, 3, 1, 5, 4, 2),
                19: (0, 2, 1, 3, 4, 5),
                24: (0, 2, 1, 5, 4, 3),
                25: (0, 1, 2, 3, 4, 5),
                30: (0, 1, 2, 5, 4, 3),
                4: (0, 5, 1, 2, 4, 3),
                14: (0, 5, 1, 3, 4, 2),
                5: (0, 4, 1, 2, 5, 3),
                15: (0, 4, 1, 3, 5, 2),
                7: (0, 3, 1, 2, 5, 4),
                17: (0, 3, 1, 4, 5, 2),
                20: (0, 2, 1, 3, 5, 4),
                23: (0, 2, 1, 4, 5, 3),
                26: (0, 1, 2, 3, 5, 4),
                29: (0, 1, 2, 4, 5, 3),
                10: (0, 5, 1, 4, 2, 3),
                8: (0, 5, 1, 3, 2, 4),
                11: (0, 4, 1, 5, 2, 3),
                9: (0, 4, 1, 3, 2, 5),
                13: (0, 3, 1, 5, 2, 4),
                12: (0, 3, 1, 4, 2, 5),
                22: (0, 2, 1, 5, 3, 4),
                21: (0, 2, 1, 4, 3, 5),
                28: (0, 1, 2, 5, 3, 4),
                27: (0, 1, 2, 4, 3, 5),
            }

            order = permutation_atom_order_dict[perm]
            neigh_atoms = tuple([neighbors[i] for i in order])
            oct_atoms = (id_atom_map[atom_idx], *neigh_atoms)
            assert len(oct_atoms) == 7
            atom_stereo = Octahedral(oct_atoms, 1)
            graph.set_atom_stereo(atom_stereo)

    for bond in (
        b
        for b in rdmol.GetBonds()
        if b.GetIsConjugated()
        or b.GetBondType() == Chem.rdchem.BondType.DOUBLE
        or b.GetStereo()
        in (Chem.BondStereo.STEREOATROPCW, Chem.BondStereo.STEREOATROPCCW)
    ):
        begin_end_idx: tuple[int, int] = (bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())

        neighbors_begin: list[int] = [
            atom.GetIdx()
            for atom in rdmol.GetAtomWithIdx(begin_end_idx[0]).GetNeighbors()
            if atom.GetIdx() != begin_end_idx[1]
        ]

        neighbors_end: list[int] = [
            atom.GetIdx()
            for atom in rdmol.GetAtomWithIdx(begin_end_idx[1]).GetNeighbors()
            if atom.GetIdx() != begin_end_idx[0]
        ]

        if len({*neighbors_begin, *neighbors_end}) != 4:
            continue
        # TODO: how to deal with double bonds in strained structures?
        # cyclopropane ?

        if len(neighbors_begin) != 2 or len(neighbors_end) != 2:
            continue
        # TODO: how to deal with imines?

        elif bond.GetStereo() in (
            Chem.BondStereo.STEREOATROPCW,
            Chem.BondStereo.STEREOATROPCCW,
        ):
            raise NotImplementedError("Atropisomerism is not implemented yet. ")

        elif (
            bond.GetBondType() == Chem.rdchem.BondType.DOUBLE
            and [a for a in bond.GetStereoAtoms()] != []
        ):
            if bond.GetStereo() == Chem.BondStereo.STEREONONE:
                bond_atoms_idx = (
                    (
                        *neighbors_begin,
                        begin_end_idx[0],
                        begin_end_idx[1],
                        *neighbors_end,
                    )
                )
                
                bond_atoms = tuple([id_atom_map[i] for i in bond_atoms_idx])
                assert len(bond_atoms) == 6
                stereo = PlanarBond(bond_atoms, None)
                invert = None
            else:
                if bond.GetStereo() == Chem.BondStereo.STEREOZ:
                    invert = False
                elif bond.GetStereo() == Chem.BondStereo.STEREOE:
                    invert = True
                else:
                    raise RuntimeError("Unknown Stereo")

                stereo_atoms = [a for a in bond.GetStereoAtoms()]

                if (
                    stereo_atoms[0] in neighbors_begin
                    and stereo_atoms[1] in neighbors_end
                ):
                    bond_atoms_idx = (
                        stereo_atoms[0],
                        *[n for n in neighbors_begin if n != stereo_atoms[0]],
                        begin_end_idx[0],
                        begin_end_idx[1],
                        stereo_atoms[1],
                        *[n for n in neighbors_end if n != stereo_atoms[1]],
                    )

                    bond_atoms = tuple([id_atom_map[a] for a in bond_atoms_idx])

                    # raise Exception(bond_atoms_idx)

                elif (
                    stereo_atoms[0] in neighbors_end
                    and stereo_atoms[1] in neighbors_begin
                ):
                    bond_atoms_idx = (
                        stereo_atoms[0],
                        *[n for n in neighbors_end if n != stereo_atoms[0]],
                        begin_end_idx[1],
                        begin_end_idx[0],
                        stereo_atoms[1],
                        *[n for n in neighbors_begin if n != stereo_atoms[1]],
                    )

                    bond_atoms = tuple([id_atom_map[a] for a in bond_atoms_idx])
                else:
                    raise RuntimeError("Stereo Atoms not neighbors")

                if invert is True:
                    inverted_atoms = tuple([bond_atoms[i] for i in (1, 0, 2, 3, 4, 5)])
                    assert len(inverted_atoms) == 6
                    stereo = PlanarBond(inverted_atoms, 0)
                elif invert is False:
                    assert len(bond_atoms) == 6
                    stereo = PlanarBond(bond_atoms, 0)

        elif bond.GetBondType() == Chem.rdchem.BondType.AROMATIC:
            ri = rdmol.GetRingInfo()
            rings: list[set[int]] = [set(ring) for ring in ri.AtomRings()]
            stereo_atoms = [
                neighbors_begin[0],
                neighbors_begin[1],
                begin_end_idx[0],
                begin_end_idx[1],
                neighbors_end[0],
                neighbors_end[1],
            ]

            common_ring_size_db1: None | int = None
            for ring in rings:
                if all(
                    a in ring
                    for a in (
                        neighbors_begin[0],
                        begin_end_idx[0],
                        begin_end_idx[1],
                        neighbors_end[0],
                    )
                ):
                    if common_ring_size_db1 is None or len(ring) < common_ring_size_db1:
                        common_ring_size_db1 = len(ring)
                if all(
                    a in ring
                    for a in (
                        neighbors_begin[1],
                        begin_end_idx[0],
                        begin_end_idx[1],
                        neighbors_end[1],
                    )
                ):
                    if common_ring_size_db1 is None or len(ring) < common_ring_size_db1:
                        common_ring_size_db1 = len(ring)

            common_ring_size_db2: None|int = None

            for ring in rings:
                if all(
                    a in ring
                    for a in (
                        neighbors_begin[1],
                        begin_end_idx[0],
                        begin_end_idx[1],
                        neighbors_end[0],
                    )
                ):
                    if common_ring_size_db2 is None or len(ring) < common_ring_size_db2:
                        common_ring_size_db2 = len(ring)
            for ring in rings:
                if all(
                    a in ring
                    for a in (
                        neighbors_begin[0],
                        begin_end_idx[0],
                        begin_end_idx[1],
                        neighbors_end[1],
                    )
                ):
                    if common_ring_size_db2 is None or len(ring) < common_ring_size_db2:
                        common_ring_size_db2 = len(ring)

            if common_ring_size_db1 is None and common_ring_size_db2 is None:
                pb_atoms = tuple([id_atom_map[a] for a in stereo_atoms])
                assert len(pb_atoms) == 6
                stereo = PlanarBond(pb_atoms, None)
            elif common_ring_size_db1:
                pb_atoms = tuple(tuple([id_atom_map[a] for a in stereo_atoms]))
                assert len(pb_atoms) == 6
                stereo = PlanarBond(pb_atoms, parity=0)
            elif common_ring_size_db2:
                pb_atoms = tuple(tuple([id_atom_map[stereo_atoms[i]] for i in (0, 1, 2, 3, 4, 5)]))
                assert len(pb_atoms) == 6
                stereo = PlanarBond(pb_atoms, parity=0,)
            elif common_ring_size_db1 is None or common_ring_size_db2 is None:
                raise RuntimeError(
                    "Aromatic Atoms not in ring, "
                    "please check the input structure"
                )
            elif common_ring_size_db2 < common_ring_size_db1:
                pb_atoms = tuple(tuple([id_atom_map[a] for a in stereo_atoms]))
                assert len(pb_atoms) == 6
                stereo = PlanarBond(pb_atoms, parity=0)
            elif common_ring_size_db1 < common_ring_size_db2:
                pb_atoms = tuple([id_atom_map[stereo_atoms[i]] for i in (0, 1, 2, 3, 4, 5)])
                assert len(pb_atoms) == 6
                stereo = PlanarBond(pb_atoms, parity=0)
            else:
                raise RuntimeError("Aromatic Atoms not in ring")

        else:
            stereo_atoms = [
                neighbors_begin[0],
                neighbors_begin[1],
                begin_end_idx[0],
                begin_end_idx[1],
                neighbors_end[0],
                neighbors_end[1],
            ]
            pb_atoms = tuple([id_atom_map[a] for a in stereo_atoms])
            assert len(pb_atoms) == 6
            stereo = PlanarBond(pb_atoms, None)

        # raise Exception(begin_end_idx, )
        bond_atoms = [id_atom_map[i] for i in begin_end_idx]

        graph.set_bond_stereo(stereo) # type: ignore

    return graph



# This is a derived work from https://github.com/jensengroup/xyz2mol
# Jensen Group
# Jan H. Jensen Research Group of the Department of Chemistry,
# University of Copenhagen
# License: MIT License (see at end of file)
#
# This code is based on the work of DOI: 10.1002/bkcs.10334
# Yeonjoon Kim and Woo Youn Kim
# "Universal Structure Conversion Method for Organic Molecules:
# From Atomic Connectivity to Three-Dimensional Geometry"
# Bull. Korean Chem. Soc.
# 2015, Vol. 36, 1769-1777

from __future__ import annotations

import copy
import itertools
import warnings
from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import TypeVar

    from stereomolgraph import Element
    N = TypeVar('N', bound=int)

atomic_valence:dict[int, list[int]] = defaultdict(lambda: [0,1,2,3,4,5,6,7,8])
atomic_valence[1] = [1]
atomic_valence[5] = [3, 4]
atomic_valence[6] = [4]
atomic_valence[7] = [3, 4]
atomic_valence[8] = [2, 1, 3]
atomic_valence[9] = [1]
atomic_valence[14] = [4]
atomic_valence[15] = [5, 3]  # [5,4,3]
atomic_valence[16] = [6, 3, 2]  # [6,4,2]
atomic_valence[17] = [1]
atomic_valence[32] = [4]
atomic_valence[35] = [1]
atomic_valence[53] = [1]
atomic_valence[78] = [2, 4]

atomic_valence_electrons: dict[int, int] = {}
atomic_valence_electrons[1] = 1
atomic_valence_electrons[5] = 3
atomic_valence_electrons[6] = 4
atomic_valence_electrons[7] = 5
atomic_valence_electrons[8] = 6
atomic_valence_electrons[9] = 7
atomic_valence_electrons[14] = 4
atomic_valence_electrons[15] = 5
atomic_valence_electrons[16] = 6
atomic_valence_electrons[17] = 7
atomic_valence_electrons[32] = 4
atomic_valence_electrons[35] = 7
atomic_valence_electrons[53] = 7
atomic_valence_electrons[78] = 10


def connectivity2bond_orders(
    atom_types: Sequence[Element],
    connectivity_matrix: np.ndarray[tuple[N, N], np.dtype[np.integer]],
    allow_charged_fragments:bool=False,
    charge: int = 0,
) -> tuple[np.ndarray[tuple[N, N], np.dtype[np.int8]],
           list[int],
           list[int]]:
    """
    Calculates Bond orders from atom connectivity.
    Bond orders can be assigned automatically using the algorithm from
    DOI: 10.1002/bkcs.10334
    Yeonjoon Kim and Woo Youn Kim
    "Universal Structure Conversion Method for Organic Molecules:
    From Atomic Connectivity to Three-Dimensional Geometry"
    Bull. Korean Chem. Soc.
    2015, Vol. 36, 1769-1777

    :param atom_types: list of atom types
    :param connectivity_matrix: Adjacency matrix. Has to be symmetric square
                                matrix. Behaveour is not defined for values
                                other than 1 and 0.
    :param allow_charged_fragments: If false radicals are formed and if True
                                ions are preferred, defaults to False.
                                bond_orders has to be set to true to be able to
                                assign charges to fragments.
    :param charge: charge of the whole molecule, defaults to 0.
    :return: bond_order_matrix, atomic_charges, atomic_valence_electrons
    """
    con_mat = np.array(connectivity_matrix, dtype=int)
    assert len (atom_types) == np.shape(connectivity_matrix)[0], (
        "atom_types and connectivity_matrix have to be of the same length"
    )
    atom_nrs = [elem.atomic_nr for elem in atom_types]
    charges = [0] * len(atom_nrs)
    unpaired_electrons = [0] * len(atom_nrs)

    # convert AC matrix to bond order (BO) matrix
    BO_matrix, atomic_valence_electrons = _AC2BO(
            con_mat,
            atom_nrs,
            charge,
            allow_charged_fragments=allow_charged_fragments,
        )
    
    BO_valences = [sum(row) for row in BO_matrix]

    # set atomic charges
    if allow_charged_fragments:
        mol_charge = 0
        for i, atom_nr in enumerate(atom_nrs):
            atom_charge = _get_atomic_charge(atom_nr,
                                   atomic_valence_electrons=atomic_valence_electrons[atom_nr],
                                   BO_valence=BO_valences[i])
            mol_charge += atom_charge
            if atom_nr == 6:
                number_of_single_bonds_to_C = list(BO_matrix[i]).count(1)
                if number_of_single_bonds_to_C == 2 and BO_valences[i] == 2:
                    mol_charge += 1
                    atom_charge = 0
                if number_of_single_bonds_to_C == 3 and charge + 1 < charge:
                    mol_charge += 2
                    atom_charge = 1
            if (abs(charge) > 0):
                charges[i] = atom_charge

    # set atomic radicals
    for i, atom_nr in enumerate(atom_nrs):
        atom_charge = _get_atomic_charge(atom_nr,
                                   atomic_valence_electrons=atomic_valence_electrons[atom_nr],
                                   BO_valence=BO_valences[i])
        if abs(atom_charge) > 0:
            unpaired_electrons[i] = abs(int(atom_charge))

    return BO_matrix, charges, unpaired_electrons

def _AC2BO(AC: np.ndarray[tuple[N, N], np.dtype[np.int8]],
           atom_nrs: list[int], charge: int, allow_charged_fragments:bool=True
           ) -> tuple[np.ndarray[tuple[N, N], np.dtype[np.int8]],
                      dict[int, int]]:
    """

    implemenation of algorithm shown in Figure 2

    UA: unsaturated atoms

    DU: degree of unsaturation (u matrix in Figure)

    best_BO: Bcurr in Figure

    """

    # make a list of valences, e.g. for CO: [[4],[2,1]]
    valences_list_of_lists:list[list[int]] = []
    AC_valence: list[int] = list(AC.sum(axis=1))

    for i, (atomicNum, valence) in enumerate(zip(atom_nrs, AC_valence)):
        # valence can't be smaller than number of neighbourgs
        possible_valence: list[int] = [
            x for x in atomic_valence[atomicNum] if x >= valence
        ]
        if not possible_valence:
            warnings.warn(
                f"Valence of atom {i},is {valence}, which bigger than allowed "
                f"max {max(atomic_valence[atomicNum])}. Continuing"
            )
            # sys.exit()
        valences_list_of_lists.append(possible_valence)

    # convert [[4],[2,1]] to [[4,2],[4,1]]
    valences_list = itertools.product(*valences_list_of_lists)

    best_BO = AC.copy()

    for valences in valences_list:
        UA, DU_from_AC = _get_UA(valences, AC_valence)

        check_len = len(UA) == 0
        if check_len:
            check_bo = _BO_is_OK(
                AC,
                AC,
                charge,
                DU_from_AC,
                atomic_valence_electrons,
                atom_nrs,
                valences,
                allow_charged_fragments=allow_charged_fragments,
            )
        else:
            check_bo = None

        if check_len and check_bo:
            return AC, atomic_valence_electrons

        UA_pairs_list = _get_UA_pairs(UA, AC)
        for UA_pairs in UA_pairs_list:
            BO = _get_BO(
                AC, UA, DU_from_AC, valences, UA_pairs
            )
            status = _BO_is_OK(
                BO,
                AC,
                charge,
                DU_from_AC,
                atomic_valence_electrons,
                atom_nrs,
                valences,
                allow_charged_fragments=allow_charged_fragments,
            )
            charge_OK = _charge_is_OK(
                BO,
                AC,
                charge,
                DU_from_AC,
                atomic_valence_electrons,
                atom_nrs,
                valences,
                allow_charged_fragments=allow_charged_fragments,
            )

            if status:
                return BO, atomic_valence_electrons
            elif (
                BO.sum() >= best_BO.sum()
                and _valences_not_too_large(BO, valences)
                and charge_OK
            ):
                best_BO = BO.copy()

    return best_BO, atomic_valence_electrons

def _get_UA(maxValence_list: Sequence[int], valence_list: list[int]
            ) -> tuple[list[int], list[int]]:
    UA: list[int] = []
    DU: list[int] = []
    for i, (maxValence, valence) in enumerate(
        zip(maxValence_list, valence_list)
    ):
        if not maxValence - valence > 0:
            continue
        UA.append(i)
        DU.append(maxValence - valence)
    return UA, DU

def _get_BO(AC: np.ndarray[tuple[N, N]],
            UA: Sequence[int],
            DU: Sequence[int],
            valences: Sequence[int],
            UA_pairs: tuple[tuple[int, int], ...]):
    BO = AC.copy()
    DU_save = []

    while DU_save != DU:
        for i, j in UA_pairs:
            BO[i, j] += 1
            BO[j, i] += 1

        BO_valence = list(BO.sum(axis=1))
        DU_save = copy.copy(DU)
        UA, DU = _get_UA(valences, BO_valence) # type: ignore[assignment]
        UA_pairs = _get_UA_pairs(UA, AC)[0]

    return BO

def _valences_not_too_large(BO: np.ndarray[tuple[N, N], np.dtype[np.int8]], valences: Sequence[int]) -> bool:
    number_of_bonds_list = BO.sum(axis=1)
    for valence, number_of_bonds in zip(valences, number_of_bonds_list):
        if number_of_bonds > valence:
            return False

    return True

def _BO_is_OK(
    BO : np.ndarray[tuple[N, N], np.dtype[np.int8]],
    AC : np.ndarray[tuple[N, N], np.dtype[np.int8]],
    charge : int,
    DU: list[int],
    atomic_valence_electrons: dict[int, int],
    atom_nrs: list[int],
    valences: Sequence[int],
    allow_charged_fragments:bool=True,
) -> bool:

    if not _valences_not_too_large(BO, valences):
        return False

    check_sum = (BO - AC).sum() == sum(DU)
    check_charge = _charge_is_OK(
        BO,
        AC,
        charge,
        DU,
        atomic_valence_electrons,
        atom_nrs,
        valences,
        allow_charged_fragments,
    )

    if check_charge and check_sum:
        return True

    return False

def _charge_is_OK(
    BO: np.ndarray[tuple[N, N], np.dtype[np.int8]],
    AC: np.ndarray[tuple[N, N], np.dtype[np.int8]],
    charge: int,
    DU: list[int],
    atomic_valence_electrons: dict[int, int],
    atom_nrs: list[int],
    valences: Sequence[int],
    allow_charged_fragments:bool=True,
) -> bool:
    # total charge
    q_tot = 0

    # charge fragment list
    q_list: list[int] = []

    if allow_charged_fragments:
        BO_valences = list(BO.sum(axis=1))
        for i, atom in enumerate(atom_nrs):
            q: int = _get_atomic_charge(
                atom, atomic_valence_electrons[atom], BO_valences[i]
            )
            q_tot += q
            if atom == 6:
                number_of_single_bonds_to_C = list(BO[i, :]).count(1)
                if number_of_single_bonds_to_C == 2 and BO_valences[i] == 2:
                    q_tot += 1
                    q = 2
                if number_of_single_bonds_to_C == 3 and q_tot + 1 < charge:
                    q_tot += 2
                    q = 1

            if q != 0:
                q_list.append(q)

    return charge == q_tot

def _get_UA_pairs(UA: Sequence[int], AC: np.ndarray[tuple[N, N], np.dtype[np.int8]]
                   ) -> list[tuple[()]] | list[tuple[tuple[int, int], ...]]:
    bonds = _get_bonds(UA, AC)

    if len(bonds) == 0:
        return [()]

    max_atoms_in_combo = 0
    UA_pairs = [()]
    for combo in list(itertools.combinations(bonds, int(len(UA) / 2))):
        flat_list = [item for sublist in combo for item in sublist]
        atoms_in_combo = len(set(flat_list))
        if atoms_in_combo > max_atoms_in_combo:
            max_atoms_in_combo = atoms_in_combo
            UA_pairs = [combo]

        elif atoms_in_combo == max_atoms_in_combo:
            UA_pairs.append(combo) # type: ignore[assignment]

    return UA_pairs

def _get_bonds(UA: Sequence[int], AC: np.ndarray[tuple[N, N], np.dtype[np.int8]]) -> list[tuple[int, int]]:
    bonds: list[tuple[int, int]] = []

    for k, i in enumerate(UA):
        for j in UA[k + 1 :]:
            if AC[i, j] == 1:
                to_append = tuple(sorted([i, j]))
                assert len(to_append) == 2
                bonds.append(to_append)

    return bonds

def _get_atomic_charge(atom: int, atomic_valence_electrons: int, BO_valence: int) -> int:
    if atom == 1:
        charge = 1 - BO_valence
    elif atom == 5:
        charge = 3 - BO_valence
    elif atom == 15 and BO_valence == 5:
        charge = 0
    elif atom == 16 and BO_valence == 6:
        charge = 0
    else:
        charge = atomic_valence_electrons - 8 + BO_valence

    return charge



# MIT License

# Copyright (c) 2018 Jensen Group

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import sys
from pathlib import Path

import nbformat
import pytest
from nbclient import NotebookClient

EXAMPLES_FOLDER = Path(__file__).parent.parent / "examples"

NOTEBOOKS = [
    "compare_molecules.ipynb",
    "phosgenation.ipynb",
    #"werner.ipynb",
    "visualization.ipynb",
]


@pytest.mark.example
@pytest.mark.parametrize("notebook", NOTEBOOKS)
def test_notebook(notebook):
    notebook_path = EXAMPLES_FOLDER / notebook

    with Path(notebook_path).open() as f:
        executed_nb = nbformat.read(f, as_version=4)

    client = NotebookClient(
        executed_nb,
        timeout=600,
        kernel_name="python",
        kernel_path=Path(sys.executable).resolve(),
    )
    client.execute()

    with Path(notebook_path).open() as f:
        nb = nbformat.read(f, as_version=4)

    for ex_cell, cell in zip(executed_nb.cells, nb.cells):
        assert ex_cell.cell_type == cell.cell_type

        if ex_cell.cell_type == "code":
            assert ex_cell.source == cell.source

            for ex_output, output in zip(ex_cell.outputs, cell.outputs):
                assert ex_output.output_type == output.output_type

                if ex_output.output_type == "stream":
                    assert ex_output.name == output.name
                    assert ex_output.text == output.text

                elif ex_output.output_type == "execute_result":
                    assert ex_output.data == output.data
                    assert ex_output.metadata == output.metadata

                elif ex_output.output_type == "display_data":
                    assert ex_output.data == output.data
                    assert ex_output.metadata == output.metadata

                elif ex_output.output_type == "error":
                    assert ex_output.ename == output.ename
                    assert ex_output.evalue == output.evalue
